# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the curve_of_growth module.
"""

import astropy.units as u
import numpy as np
import pytest
from astropy.utils.exceptions import AstropyUserWarning
from numpy.testing import assert_allclose, assert_equal

from photutils.aperture import (CircularAperture, EllipticalAperture,
                                RectangularAperture)
from photutils.profiles import (CurveOfGrowth, EllipticalCurveOfGrowth,
                                EnsquaredCurveOfGrowth, ProfileBase)
from photutils.utils._optional_deps import HAS_MATPLOTLIB


class TestCurveOfGrowth:
    def test_basic(self, profile_data):
        """
        Test basic CurveOfGrowth functionality.
        """
        xycen, data, _, _ = profile_data

        radii = np.arange(1, 37)
        cg1 = CurveOfGrowth(data, xycen, radii, error=None, mask=None)

        assert_equal(cg1.radius, radii)
        assert cg1.area.shape == (36,)
        assert cg1.profile.shape == (36,)
        assert cg1.profile_error.shape == (0,)
        assert_allclose(cg1.area[0], np.pi)

        assert len(cg1.apertures) == 36
        assert isinstance(cg1.apertures[0], CircularAperture)

        radii = np.arange(1, 36)
        cg2 = CurveOfGrowth(data, xycen, radii, error=None, mask=None)
        assert cg2.area[0] > 0.0
        assert isinstance(cg2.apertures[0], CircularAperture)

    def test_units(self, profile_data):
        """
        Test CurveOfGrowth with units.
        """
        xycen, data, error, _ = profile_data

        radii = np.arange(1, 36)
        unit = u.Jy
        cg1 = CurveOfGrowth(data << unit, xycen, radii,
                            error=error << unit, mask=None)

        assert cg1.profile.unit == unit
        assert cg1.profile_error.unit == unit

        match = 'must all have the same units'
        with pytest.raises(ValueError, match=match):
            CurveOfGrowth(data << unit, xycen, radii, error=error, mask=None)

    def test_error(self, profile_data):
        """
        Test CurveOfGrowth with error array.
        """
        xycen, data, error, _ = profile_data

        radii = np.arange(1, 36)
        cg1 = CurveOfGrowth(data, xycen, radii, error=error, mask=None)

        assert cg1.profile.shape == (35,)
        assert cg1.profile_error.shape == (35,)

    def test_mask(self, profile_data):
        """
        Test CurveOfGrowth with a mask.
        """
        xycen, data, error, mask = profile_data

        radii = np.arange(1, 36)
        cg1 = CurveOfGrowth(data, xycen, radii, error=error, mask=None)
        cg2 = CurveOfGrowth(data, xycen, radii, error=error, mask=mask)

        assert cg1.profile.sum() > cg2.profile.sum()
        assert np.sum(cg1.profile_error**2) > np.sum(cg2.profile_error**2)

    def test_normalize(self, profile_data):
        """
        Test CurveOfGrowth normalize and unnormalize methods.
        """
        xycen, data, _, _ = profile_data

        radii = np.arange(1, 36)
        cg1 = CurveOfGrowth(data, xycen, radii, error=None, mask=None)
        cg2 = CurveOfGrowth(data, xycen, radii, error=None, mask=None)

        profile1 = cg1.profile
        cg1.normalize()
        profile2 = cg1.profile
        assert np.mean(profile2) < np.mean(profile1)

        cg1.unnormalize()
        assert_allclose(cg1.profile, cg2.profile)

        cg1.normalize(method='sum')
        cg1.normalize(method='max')
        cg1.unnormalize()
        assert_allclose(cg1.profile, cg2.profile)

        cg1.normalize(method='max')
        cg1.normalize(method='sum')
        cg1.normalize(method='max')
        cg1.normalize(method='max')
        cg1.unnormalize()
        assert_allclose(cg1.profile, cg2.profile)

        cg1.normalize(method='sum')
        profile3 = cg1.profile
        assert np.mean(profile3) < np.mean(profile1)

        cg1.unnormalize()
        assert_allclose(cg1.profile, cg2.profile)

        match = "invalid method, must be 'max' or 'sum'"
        with pytest.raises(ValueError, match=match):
            cg1.normalize(method='invalid')

        cg1.__dict__['profile'] -= np.nanmax(cg1.__dict__['profile'])
        match = 'The profile cannot be normalized'
        with pytest.warns(AstropyUserWarning, match=match):
            cg1.normalize(method='max')

    def test_interp(self, profile_data):
        """
        Test CurveOfGrowth encircled energy interpolation methods.
        """
        xycen, data, _, _ = profile_data
        radii = np.arange(1, 36)
        cg1 = CurveOfGrowth(data, xycen, radii, error=None, mask=None)
        cg1.normalize()
        ee_radii = np.array([0, 5, 10, 20, 25, 50], dtype=float)
        ee_vals = cg1.calc_ee_at_radius(ee_radii)
        ee_expected = np.array([np.nan, 0.1176754, 0.39409357, 0.86635049,
                                0.95805792, np.nan])
        assert_allclose(ee_vals, ee_expected, rtol=1e-6)

        rr = cg1.calc_radius_at_ee(ee_vals)
        ee_radii[[0, -1]] = np.nan
        assert_allclose(rr, ee_radii, rtol=1e-6)

        radii = np.linspace(0.1, 36, 200)
        cg1 = CurveOfGrowth(data, xycen, radii, error=None, mask=None,
                            method='center')
        ee_vals = cg1.calc_ee_at_radius(ee_radii)
        match = 'The curve-of-growth profile is not monotonically increasing'
        with pytest.raises(ValueError, match=match):
            cg1.calc_radius_at_ee(ee_vals)

    def test_interp_nonmonotonic_start(self, profile_data):
        """
        Test that `calc_radius_at_ee` raises ValueError when the profile
        is non-monotonic at the very first point (covers the len(radius)
        < 2 branch).
        """
        xycen, data, _, _ = profile_data
        radii = np.arange(1, 36)
        cg1 = CurveOfGrowth(data, xycen, radii, error=None, mask=None)

        # Force non-monotonicity at the first point: diff[0] = 0
        # (non-positive) so idx=0 and radius[0:0] has fewer than 2
        # elements
        profile = cg1.profile.copy()
        profile[0] = profile[1]
        cg1.__dict__['profile'] = profile

        match = 'The curve-of-growth profile is not monotonically increasing'
        with pytest.raises(ValueError, match=match):
            cg1.calc_radius_at_ee(np.array([0.5]))

    def test_trim_to_monotonic_nan(self):
        """
        Test that `_trim_to_monotonic` keeps only the leading finite
        segment when NaN values are present (covers the NaN-trimming
        branch in core.py).
        """
        xarr = np.array([1.0, 2.0, 3.0, 4.0, 5.0])
        profile = np.array([1.0, 3.0, np.nan, 5.0, 7.0])
        xarr_out, profile_out = ProfileBase._trim_to_monotonic(
            xarr, profile, 'test')
        assert_equal(xarr_out, np.array([1.0, 2.0]))
        assert_equal(profile_out, np.array([1.0, 3.0]))

    def test_inputs(self, profile_data):
        """
        Test CurveOfGrowth input validation.
        """
        xycen, data, error, _ = profile_data

        match = 'radii must be > 0'
        radii = np.arange(10)
        with pytest.raises(ValueError, match=match):
            CurveOfGrowth(data, xycen, radii, error=None, mask=None)

        match = 'radii must be a 1D array and have at least two values'
        with pytest.raises(ValueError, match=match):
            CurveOfGrowth(data, xycen, [1], error=None, mask=None)
        with pytest.raises(ValueError, match=match):
            CurveOfGrowth(data, xycen, np.arange(1, 7).reshape(2, 3),
                          error=None, mask=None)

        match = 'radii must be strictly increasing'
        radii = np.arange(1, 10)[::-1]
        with pytest.raises(ValueError, match=match):
            CurveOfGrowth(data, xycen, radii, error=None, mask=None)

        unit1 = u.Jy
        unit2 = u.km
        radii = np.arange(1, 36)
        match = 'must all have the same units'
        with pytest.raises(ValueError, match=match):
            CurveOfGrowth(data << unit1, xycen, radii,
                          error=error << unit2)

    def test_no_mutation(self, profile_data):
        """
        Test that input data, error, mask, and radii arrays are not
        mutated by CurveOfGrowth.
        """
        xycen, data, error, mask = profile_data

        # Introduce a NaN to trigger the badmask / mask-merge code path.
        # Use a pixel outside the fixture's masked region (mask[:39, :50]).
        data2 = data.copy()
        data2[50, 70] = np.nan
        mask2 = mask.copy()
        data2_orig = data2.copy()
        error_orig = error.copy()
        mask2_orig = mask2.copy()
        radii = np.arange(1, 36)
        radii_orig = radii.copy()

        match = 'Input data contains non-finite values'
        with pytest.warns(AstropyUserWarning, match=match):
            CurveOfGrowth(data2, xycen, radii, error=error, mask=mask2)

        assert_equal(data2, data2_orig)
        assert_equal(error, error_orig)
        assert_equal(mask2, mask2_orig)
        assert_equal(radii, radii_orig)

    @pytest.mark.skipif(not HAS_MATPLOTLIB, reason='matplotlib is required')
    def test_plot(self, profile_data):
        """
        Test CurveOfGrowth plot methods.
        """
        xycen, data, error, _ = profile_data

        radii = np.arange(1, 36)
        cg1 = CurveOfGrowth(data, xycen, radii, error=None, mask=None)
        cg1.plot()
        match = 'Errors were not input'
        with pytest.warns(AstropyUserWarning, match=match):
            cg1.plot_error()

        cg2 = CurveOfGrowth(data, xycen, radii, error=error, mask=None)
        cg2.plot()
        pc1 = cg2.plot_error()
        assert_allclose(pc1.get_facecolor(), [[0.5, 0.5, 0.5, 0.3]])
        pc2 = cg2.plot_error(facecolor='blue')
        assert_allclose(pc2.get_facecolor(), [[0, 0, 1, 1]])

        unit = u.Jy
        cg3 = CurveOfGrowth(data << unit, xycen, radii,
                            error=error << unit, mask=None)
        cg3.plot()
        cg3.plot_error()

    def test_all_masked(self, profile_data):
        """
        Test CurveOfGrowth with all data masked.

        When every pixel is masked the profile should be all zero
        (zero flux in every aperture).
        """
        xycen, data, _, _ = profile_data

        all_mask = np.ones(data.shape, dtype=bool)
        radii = np.arange(1, 36)
        cg = CurveOfGrowth(data, xycen, radii, error=None, mask=all_mask)
        assert_allclose(cg.profile, 0.0)

    @pytest.mark.skipif(not HAS_MATPLOTLIB, reason='matplotlib is required')
    def test_plot_error_none(self, profile_data):
        """
        Test that ``plot_error()`` returns ``None`` when no errors were
        input.
        """
        xycen, data, _, _ = profile_data

        radii = np.arange(1, 36)
        cg = CurveOfGrowth(data, xycen, radii, error=None)
        match = 'Errors were not input'
        with pytest.warns(AstropyUserWarning, match=match):
            result = cg.plot_error()
        assert result is None

    def test_repr(self, profile_data):
        """
        Test __repr__ output format.
        """
        xycen, data, _, _ = profile_data

        radii = np.arange(1, 36)
        cg = CurveOfGrowth(data, xycen, radii)
        r = repr(cg)
        assert 'CurveOfGrowth' in r
        assert f'xycen={xycen}' in r
        assert f'n_radii={len(radii)}' in r
        assert 'normalized=False' in r


class TestEnsquaredCurveOfGrowth:
    def test_basic(self, profile_data):
        """
        Test basic EnsquaredCurveOfGrowth functionality.
        """
        xycen, data, _, _ = profile_data

        half_sizes = np.arange(1, 37)
        ecg1 = EnsquaredCurveOfGrowth(data, xycen, half_sizes, error=None,
                                      mask=None)

        assert_equal(ecg1.half_size, half_sizes)
        assert_equal(ecg1.radius, half_sizes)
        assert ecg1.area.shape == (36,)
        assert ecg1.profile.shape == (36,)
        assert ecg1.profile_error.shape == (0,)
        assert_allclose(ecg1.area[0], 4.0)  # 2x2 square

        assert len(ecg1.apertures) == 36
        assert isinstance(ecg1.apertures[0], RectangularAperture)

        half_sizes = np.arange(1, 36)
        ecg2 = EnsquaredCurveOfGrowth(data, xycen, half_sizes, error=None,
                                      mask=None)
        assert ecg2.area[0] > 0.0
        assert isinstance(ecg2.apertures[0], RectangularAperture)

    def test_units(self, profile_data):
        """
        Test EnsquaredCurveOfGrowth with units.
        """
        xycen, data, error, _ = profile_data

        half_sizes = np.arange(1, 36)
        unit = u.Jy
        ecg1 = EnsquaredCurveOfGrowth(data << unit, xycen, half_sizes,
                                      error=error << unit, mask=None)

        assert ecg1.profile.unit == unit
        assert ecg1.profile_error.unit == unit

        match = 'must all have the same units'
        with pytest.raises(ValueError, match=match):
            EnsquaredCurveOfGrowth(data << unit, xycen, half_sizes,
                                   error=error, mask=None)

    def test_error(self, profile_data):
        """
        Test EnsquaredCurveOfGrowth with error array.
        """
        xycen, data, error, _ = profile_data

        half_sizes = np.arange(1, 36)
        ecg1 = EnsquaredCurveOfGrowth(data, xycen, half_sizes, error=error,
                                      mask=None)

        assert ecg1.profile.shape == (35,)
        assert ecg1.profile_error.shape == (35,)

    def test_mask(self, profile_data):
        """
        Test EnsquaredCurveOfGrowth with a mask.
        """
        xycen, data, error, mask = profile_data

        half_sizes = np.arange(1, 36)
        ecg1 = EnsquaredCurveOfGrowth(data, xycen, half_sizes, error=error,
                                      mask=None)
        ecg2 = EnsquaredCurveOfGrowth(data, xycen, half_sizes, error=error,
                                      mask=mask)

        assert ecg1.profile.sum() > ecg2.profile.sum()
        assert np.sum(ecg1.profile_error**2) > np.sum(ecg2.profile_error**2)

    def test_normalize(self, profile_data):
        """
        Test EnsquaredCurveOfGrowth normalize and unnormalize methods.
        """
        xycen, data, _, _ = profile_data

        half_sizes = np.arange(1, 36)
        ecg1 = EnsquaredCurveOfGrowth(data, xycen, half_sizes, error=None,
                                      mask=None)
        ecg2 = EnsquaredCurveOfGrowth(data, xycen, half_sizes, error=None,
                                      mask=None)

        profile1 = ecg1.profile
        ecg1.normalize()
        profile2 = ecg1.profile
        assert np.mean(profile2) < np.mean(profile1)

        ecg1.unnormalize()
        assert_allclose(ecg1.profile, ecg2.profile)

        ecg1.normalize(method='sum')
        ecg1.normalize(method='max')
        ecg1.unnormalize()
        assert_allclose(ecg1.profile, ecg2.profile)

        ecg1.normalize(method='max')
        ecg1.normalize(method='sum')
        ecg1.normalize(method='max')
        ecg1.normalize(method='max')
        ecg1.unnormalize()
        assert_allclose(ecg1.profile, ecg2.profile)

        ecg1.normalize(method='sum')
        profile3 = ecg1.profile
        assert np.mean(profile3) < np.mean(profile1)

        ecg1.unnormalize()
        assert_allclose(ecg1.profile, ecg2.profile)

        match = "invalid method, must be 'max' or 'sum'"
        with pytest.raises(ValueError, match=match):
            ecg1.normalize(method='invalid')

        ecg1.__dict__['profile'] -= np.nanmax(ecg1.__dict__['profile'])
        match = 'The profile cannot be normalized'
        with pytest.warns(AstropyUserWarning, match=match):
            ecg1.normalize(method='max')

    def test_interp(self, profile_data):
        """
        Test EnsquaredCurveOfGrowth encircled energy interpolation
        methods.
        """
        xycen, data, _, _ = profile_data
        half_sizes = np.arange(1, 36)
        ecg1 = EnsquaredCurveOfGrowth(data, xycen, half_sizes, error=None,
                                      mask=None)
        ecg1.normalize()
        ee_half_sizes = np.array([0, 5, 10, 20, 25, 50], dtype=float)
        ee_vals = ecg1.calc_ee_at_half_size(ee_half_sizes)
        half_sizes_back = ecg1.calc_half_size_at_ee(ee_vals)
        ee_half_sizes[[0, -1]] = np.nan
        assert_allclose(half_sizes_back, ee_half_sizes, rtol=1e-6)

        half_sizes = np.linspace(0.1, 36, 200)
        ecg1 = EnsquaredCurveOfGrowth(data, xycen, half_sizes, error=None,
                                      mask=None, method='center')
        ee_vals = ecg1.calc_ee_at_half_size(ee_half_sizes)
        match = 'The ensquared curve-of-growth profile is not monotonically'
        with pytest.raises(ValueError, match=match):
            ecg1.calc_half_size_at_ee(ee_vals)

    def test_interp_nonmonotonic_start(self, profile_data):
        """
        Test that `calc_half_size_at_ee` raises ValueError when the
        profile is non-monotonic at the very first point (covers the
        len(half_size) < 2 branch).
        """
        xycen, data, _, _ = profile_data
        half_sizes = np.arange(1, 36)
        ecg1 = EnsquaredCurveOfGrowth(data, xycen, half_sizes, error=None,
                                      mask=None)
        # Force non-monotonicity at the first point
        profile = ecg1.profile.copy()
        profile[0] = profile[1]
        ecg1.__dict__['profile'] = profile

        match = 'The ensquared curve-of-growth profile is not monotonically'
        with pytest.raises(ValueError, match=match):
            ecg1.calc_half_size_at_ee(np.array([0.5]))

    def test_inputs(self, profile_data):
        """
        Test EnsquaredCurveOfGrowth input validation.
        """
        xycen, data, error, _ = profile_data

        match = 'half_sizes must be > 0'
        half_sizes = np.arange(10)
        with pytest.raises(ValueError, match=match):
            EnsquaredCurveOfGrowth(data, xycen, half_sizes, error=None,
                                   mask=None)

        match = 'radii must be a 1D array and have at least two values'
        with pytest.raises(ValueError, match=match):
            EnsquaredCurveOfGrowth(data, xycen, [1], error=None, mask=None)
        with pytest.raises(ValueError, match=match):
            EnsquaredCurveOfGrowth(data, xycen,
                                   np.arange(1, 7).reshape(2, 3),
                                   error=None, mask=None)

        match = 'radii must be strictly increasing'
        half_sizes = np.arange(1, 10)[::-1]
        with pytest.raises(ValueError, match=match):
            EnsquaredCurveOfGrowth(data, xycen, half_sizes, error=None,
                                   mask=None)

        unit1 = u.Jy
        unit2 = u.km
        half_sizes = np.arange(1, 36)
        match = 'must all have the same units'
        with pytest.raises(ValueError, match=match):
            EnsquaredCurveOfGrowth(data << unit1, xycen, half_sizes,
                                   error=error << unit2)

    def test_no_mutation(self, profile_data):
        """
        Test that input data, error, mask, and half_sizes arrays are not
        mutated by EnsquaredCurveOfGrowth.
        """
        xycen, data, error, mask = profile_data

        data2 = data.copy()
        data2[50, 70] = np.nan
        mask2 = mask.copy()
        data2_orig = data2.copy()
        error_orig = error.copy()
        mask2_orig = mask2.copy()
        half_sizes = np.arange(1, 36)
        half_sizes_orig = half_sizes.copy()

        match = 'Input data contains non-finite values'
        with pytest.warns(AstropyUserWarning, match=match):
            EnsquaredCurveOfGrowth(data2, xycen, half_sizes, error=error,
                                   mask=mask2)

        assert_equal(data2, data2_orig)
        assert_equal(error, error_orig)
        assert_equal(mask2, mask2_orig)
        assert_equal(half_sizes, half_sizes_orig)

    @pytest.mark.skipif(not HAS_MATPLOTLIB, reason='matplotlib is required')
    def test_plot(self, profile_data):
        """
        Test EnsquaredCurveOfGrowth plot methods.
        """
        xycen, data, error, _ = profile_data

        half_sizes = np.arange(1, 36)
        ecg1 = EnsquaredCurveOfGrowth(data, xycen, half_sizes, error=None,
                                      mask=None)
        ecg1.plot()
        match = 'Errors were not input'
        with pytest.warns(AstropyUserWarning, match=match):
            ecg1.plot_error()

        ecg2 = EnsquaredCurveOfGrowth(data, xycen, half_sizes, error=error,
                                      mask=None)
        ecg2.plot()
        pc1 = ecg2.plot_error()
        assert_allclose(pc1.get_facecolor(), [[0.5, 0.5, 0.5, 0.3]])
        pc2 = ecg2.plot_error(facecolor='blue')
        assert_allclose(pc2.get_facecolor(), [[0, 0, 1, 1]])

        unit = u.Jy
        ecg3 = EnsquaredCurveOfGrowth(data << unit, xycen, half_sizes,
                                      error=error << unit, mask=None)
        ecg3.plot()
        ecg3.plot_error()

    def test_repr(self, profile_data):
        """
        Test __repr__ output format.
        """
        xycen, data, _, _ = profile_data

        half_sizes = np.arange(1, 36)
        ecg = EnsquaredCurveOfGrowth(data, xycen, half_sizes)
        r = repr(ecg)
        assert 'EnsquaredCurveOfGrowth' in r
        assert f'xycen={xycen}' in r
        assert f'n_half_sizes={len(half_sizes)}' in r
        assert 'normalized=False' in r


class TestEllipticalCurveOfGrowth:
    def test_basic(self, profile_data):
        """
        Test basic EllipticalCurveOfGrowth functionality.
        """
        xycen, data, _, _ = profile_data

        radii = np.arange(1, 37)
        ecg1 = EllipticalCurveOfGrowth(data, xycen, radii, axis_ratio=0.5,
                                       error=None, mask=None)

        assert_equal(ecg1.radius, radii)
        assert ecg1.area.shape == (36,)
        assert ecg1.profile.shape == (36,)
        assert ecg1.profile_error.shape == (0,)
        assert_allclose(ecg1.area[0], np.pi * 1.0 * 0.5)

        assert len(ecg1.apertures) == 36
        assert isinstance(ecg1.apertures[0], EllipticalAperture)

        radii = np.arange(1, 36)
        ecg2 = EllipticalCurveOfGrowth(data, xycen, radii, axis_ratio=0.5,
                                       error=None, mask=None)
        assert ecg2.area[0] > 0.0
        assert isinstance(ecg2.apertures[0], EllipticalAperture)

    def test_axis_ratio_1(self, profile_data):
        """
        Test that axis_ratio=1 gives the same result as CurveOfGrowth.
        """
        xycen, data, error, _ = profile_data
        radii = np.arange(1, 36)
        cog = CurveOfGrowth(data, xycen, radii, error=error, mask=None)
        ecg = EllipticalCurveOfGrowth(data, xycen, radii, axis_ratio=1.0,
                                      error=error, mask=None)
        assert_allclose(ecg.profile, cog.profile)
        assert_allclose(ecg.profile_error, cog.profile_error)

    def test_units(self, profile_data):
        """
        Test EllipticalCurveOfGrowth with units.
        """
        xycen, data, error, _ = profile_data

        radii = np.arange(1, 36)
        unit = u.Jy
        ecg1 = EllipticalCurveOfGrowth(data << unit, xycen, radii,
                                       axis_ratio=0.5,
                                       error=error << unit, mask=None)

        assert ecg1.profile.unit == unit
        assert ecg1.profile_error.unit == unit

        match = 'must all have the same units'
        with pytest.raises(ValueError, match=match):
            EllipticalCurveOfGrowth(data << unit, xycen, radii,
                                    axis_ratio=0.5, error=error, mask=None)

    def test_error(self, profile_data):
        """
        Test EllipticalCurveOfGrowth with error array.
        """
        xycen, data, error, _ = profile_data

        radii = np.arange(1, 36)
        ecg1 = EllipticalCurveOfGrowth(data, xycen, radii, axis_ratio=0.5,
                                       error=error, mask=None)

        assert ecg1.profile.shape == (35,)
        assert ecg1.profile_error.shape == (35,)

    def test_mask(self, profile_data):
        """
        Test EllipticalCurveOfGrowth with a mask.
        """
        xycen, data, error, mask = profile_data

        radii = np.arange(1, 36)
        ecg1 = EllipticalCurveOfGrowth(data, xycen, radii, axis_ratio=0.5,
                                       error=error, mask=None)
        ecg2 = EllipticalCurveOfGrowth(data, xycen, radii, axis_ratio=0.5,
                                       error=error, mask=mask)

        assert ecg1.profile.sum() > ecg2.profile.sum()
        assert np.sum(ecg1.profile_error**2) > np.sum(ecg2.profile_error**2)

    def test_normalize(self, profile_data):
        """
        Test EllipticalCurveOfGrowth normalize and unnormalize methods.
        """
        xycen, data, _, _ = profile_data

        radii = np.arange(1, 36)
        ecg1 = EllipticalCurveOfGrowth(data, xycen, radii, axis_ratio=0.5,
                                       error=None, mask=None)
        ecg2 = EllipticalCurveOfGrowth(data, xycen, radii, axis_ratio=0.5,
                                       error=None, mask=None)

        profile1 = ecg1.profile
        ecg1.normalize()
        profile2 = ecg1.profile
        assert np.mean(profile2) < np.mean(profile1)

        ecg1.unnormalize()
        assert_allclose(ecg1.profile, ecg2.profile)

        ecg1.normalize(method='sum')
        ecg1.normalize(method='max')
        ecg1.unnormalize()
        assert_allclose(ecg1.profile, ecg2.profile)

        ecg1.normalize(method='max')
        ecg1.normalize(method='sum')
        ecg1.normalize(method='max')
        ecg1.normalize(method='max')
        ecg1.unnormalize()
        assert_allclose(ecg1.profile, ecg2.profile)

        ecg1.normalize(method='sum')
        profile3 = ecg1.profile
        assert np.mean(profile3) < np.mean(profile1)

        ecg1.unnormalize()
        assert_allclose(ecg1.profile, ecg2.profile)

        match = "invalid method, must be 'max' or 'sum'"
        with pytest.raises(ValueError, match=match):
            ecg1.normalize(method='invalid')

        ecg1.__dict__['profile'] -= np.nanmax(ecg1.__dict__['profile'])
        match = 'The profile cannot be normalized'
        with pytest.warns(AstropyUserWarning, match=match):
            ecg1.normalize(method='max')

    def test_interp(self, profile_data):
        """
        Test EllipticalCurveOfGrowth encircled energy interpolation
        methods.
        """
        xycen, data, _, _ = profile_data
        radii = np.arange(1, 36)
        ecg1 = EllipticalCurveOfGrowth(data, xycen, radii, axis_ratio=0.5,
                                       error=None, mask=None)
        ecg1.normalize()
        ee_radii = np.array([0, 5, 10, 20, 25, 50], dtype=float)
        ee_vals = ecg1.calc_ee_at_radius(ee_radii)
        radii_back = ecg1.calc_radius_at_ee(ee_vals)
        ee_radii[[0, -1]] = np.nan
        assert_allclose(radii_back, ee_radii, rtol=1e-6)

        radii = np.linspace(0.1, 36, 200)
        ecg1 = EllipticalCurveOfGrowth(data, xycen, radii, axis_ratio=0.5,
                                       error=None, mask=None,
                                       method='center')
        ee_vals = ecg1.calc_ee_at_radius(ee_radii)
        match = 'The elliptical curve-of-growth profile is not monotonically'
        with pytest.raises(ValueError, match=match):
            ecg1.calc_radius_at_ee(ee_vals)

    def test_interp_nonmonotonic_start(self, profile_data):
        """
        Test that `calc_radius_at_ee` raises ValueError when the
        profile is non-monotonic at the very first point (covers the
        len(radius) < 2 branch).
        """
        xycen, data, _, _ = profile_data
        radii = np.arange(1, 36)
        ecg1 = EllipticalCurveOfGrowth(data, xycen, radii, axis_ratio=0.5,
                                       error=None, mask=None)
        # Force non-monotonicity at the first point
        profile = ecg1.profile.copy()
        profile[0] = profile[1]
        ecg1.__dict__['profile'] = profile

        match = 'The elliptical curve-of-growth profile is not monotonically'
        with pytest.raises(ValueError, match=match):
            ecg1.calc_radius_at_ee(np.array([0.5]))

    def test_inputs(self, profile_data):
        """
        Test EllipticalCurveOfGrowth input validation.
        """
        xycen, data, error, _ = profile_data

        match = 'radii must be > 0'
        radii = np.arange(10)
        with pytest.raises(ValueError, match=match):
            EllipticalCurveOfGrowth(data, xycen, radii, axis_ratio=0.5,
                                    error=None, mask=None)

        match = 'radii must be a 1D array and have at least two values'
        with pytest.raises(ValueError, match=match):
            EllipticalCurveOfGrowth(data, xycen, [1], axis_ratio=0.5,
                                    error=None, mask=None)
        with pytest.raises(ValueError, match=match):
            EllipticalCurveOfGrowth(data, xycen,
                                    np.arange(1, 7).reshape(2, 3),
                                    axis_ratio=0.5, error=None, mask=None)

        match = 'radii must be strictly increasing'
        radii = np.arange(1, 10)[::-1]
        with pytest.raises(ValueError, match=match):
            EllipticalCurveOfGrowth(data, xycen, radii, axis_ratio=0.5,
                                    error=None, mask=None)

        match = 'axis_ratio must be in the range 0 < axis_ratio <= 1'
        radii = np.arange(1, 36)
        with pytest.raises(ValueError, match=match):
            EllipticalCurveOfGrowth(data, xycen, radii, axis_ratio=0.0)
        with pytest.raises(ValueError, match=match):
            EllipticalCurveOfGrowth(data, xycen, radii, axis_ratio=-0.5)
        with pytest.raises(ValueError, match=match):
            EllipticalCurveOfGrowth(data, xycen, radii, axis_ratio=1.5)

        unit1 = u.Jy
        unit2 = u.km
        radii = np.arange(1, 36)
        match = 'must all have the same units'
        with pytest.raises(ValueError, match=match):
            EllipticalCurveOfGrowth(data << unit1, xycen, radii,
                                    axis_ratio=0.5,
                                    error=error << unit2)

    def test_no_mutation(self, profile_data):
        """
        Test that input data, error, mask, and radii arrays are not
        mutated by EllipticalCurveOfGrowth.
        """
        xycen, data, error, mask = profile_data

        data2 = data.copy()
        data2[50, 70] = np.nan
        mask2 = mask.copy()
        data2_orig = data2.copy()
        error_orig = error.copy()
        mask2_orig = mask2.copy()
        radii = np.arange(1, 36)
        radii_orig = radii.copy()

        match = 'Input data contains non-finite values'
        with pytest.warns(AstropyUserWarning, match=match):
            EllipticalCurveOfGrowth(data2, xycen, radii, axis_ratio=0.5,
                                    error=error, mask=mask2)

        assert_equal(data2, data2_orig)
        assert_equal(error, error_orig)
        assert_equal(mask2, mask2_orig)
        assert_equal(radii, radii_orig)

    @pytest.mark.skipif(not HAS_MATPLOTLIB, reason='matplotlib is required')
    def test_plot(self, profile_data):
        """
        Test EllipticalCurveOfGrowth plot methods.
        """
        xycen, data, error, _ = profile_data

        radii = np.arange(1, 36)
        ecg1 = EllipticalCurveOfGrowth(data, xycen, radii, axis_ratio=0.5,
                                       error=None, mask=None)
        ecg1.plot()
        match = 'Errors were not input'
        with pytest.warns(AstropyUserWarning, match=match):
            ecg1.plot_error()

        ecg2 = EllipticalCurveOfGrowth(data, xycen, radii, axis_ratio=0.5,
                                       error=error, mask=None)
        ecg2.plot()
        pc1 = ecg2.plot_error()
        assert_allclose(pc1.get_facecolor(), [[0.5, 0.5, 0.5, 0.3]])
        pc2 = ecg2.plot_error(facecolor='blue')
        assert_allclose(pc2.get_facecolor(), [[0, 0, 1, 1]])

        unit = u.Jy
        ecg3 = EllipticalCurveOfGrowth(data << unit, xycen, radii,
                                       axis_ratio=0.5,
                                       error=error << unit, mask=None)
        ecg3.plot()
        ecg3.plot_error()

    def test_repr(self, profile_data):
        """
        Test __repr__ output format.
        """
        xycen, data, _, _ = profile_data

        radii = np.arange(1, 36)
        ecg = EllipticalCurveOfGrowth(data, xycen, radii, axis_ratio=0.5)
        r = repr(ecg)
        assert 'EllipticalCurveOfGrowth' in r
        assert f'xycen={xycen}' in r
        assert f'n_radii={len(radii)}' in r
        assert 'normalized=False' in r



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Subpackage containing tools for performing point-spread-function (PSF)
photometry.
"""

from .epsf_builder import *  # noqa: F401, F403
from .epsf_stars import *  # noqa: F401, F403
from .flags import *  # noqa: F401, F403
from .functional_models import *  # noqa: F401, F403
from .gridded_models import *  # noqa: F401, F403
from .groupers import *  # noqa: F401, F403
from .image_models import *  # noqa: F401, F403
from .iterative import *  # noqa: F401, F403
from .model_helpers import *  # noqa: F401, F403
from .model_io import *  # noqa: F401, F403
from .model_plotting import *  # noqa: F401, F403
from .photometry import *  # noqa: F401, F403
from .simulation import *  # noqa: F401, F403
from .utils import *  # noqa: F401, F403



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Elliptical and elliptical-annulus apertures in both pixel and sky
coordinates.
"""

import math

import astropy.units as u
import numpy as np
from astropy.coordinates import Angle
from astropy.utils import lazyproperty

from photutils.aperture.attributes import (PixelPositions, PositiveScalar,
                                           PositiveScalarAngle, ScalarAngle,
                                           ScalarAngleOrValue,
                                           SkyCoordPositions)
from photutils.aperture.core import PixelAperture, SkyAperture
from photutils.aperture.mask import ApertureMask
from photutils.geometry import elliptical_overlap_grid
from photutils.utils._deprecation import (deprecated,
                                          deprecated_positional_kwargs)
from photutils.utils._wcs_helpers import (pixel_ellipse_to_sky_svd,
                                          sky_ellipse_to_pixel_svd)

__all__ = [
    'EllipticalAnnulus',
    'EllipticalAperture',
    'EllipticalMaskMixin',
    'SkyEllipticalAnnulus',
    'SkyEllipticalAperture',
]


@deprecated('3.0', until='4.0')
class EllipticalMaskMixin:  # pragma: no cover
    """
    Mixin class to create masks for elliptical and elliptical-annulus
    aperture objects.

    .. deprecated:: 3.0
    """

    def to_mask(self, method='exact', subpixels=5):
        """
        Return a mask for the aperture.

        Parameters
        ----------
        method : {'exact', 'center', 'subpixel'}, optional
            The method used to determine the overlap of the aperture
            on the pixel grid. Not all options are available for all
            aperture types. Note that the more precise methods are
            generally slower. The following methods are available:

            * ``'exact'`` (default):
              The exact fractional overlap of the aperture and each
              pixel is calculated. The aperture weights will contain
              values between 0 and 1.

            * ``'center'``:
              A pixel is considered to be entirely in or out of the
              aperture depending on whether its center is in or out of
              the aperture. The aperture weights will contain values
              only of 0 (out) and 1 (in).

            * ``'subpixel'``:
              A pixel is divided into subpixels (see the ``subpixels``
              keyword), each of which are considered to be entirely in
              or out of the aperture depending on whether its center is
              in or out of the aperture. If ``subpixels=1``, this method
              is equivalent to ``'center'``. The aperture weights will
              contain values between 0 and 1.

        subpixels : int, optional
            For the ``'subpixel'`` method, resample pixels by this
            factor in each dimension. That is, each pixel is divided
            into ``subpixels**2`` subpixels. This keyword is ignored
            unless ``method='subpixel'``.

        Returns
        -------
        mask : `~photutils.aperture.ApertureMask` or list of \
                `~photutils.aperture.ApertureMask`
            A mask for the aperture. If the aperture is scalar then
            a single `~photutils.aperture.ApertureMask` is returned,
            otherwise a list of `~photutils.aperture.ApertureMask` is
            returned.
        """
        use_exact, subpixels = self._translate_mask_method(method, subpixels)

        if hasattr(self, 'a'):
            a = self.a
            b = self.b
        elif hasattr(self, 'a_in'):  # annulus
            a = self.a_out
            b = self.b_out
        else:
            msg = 'Cannot determine the aperture shape'
            raise ValueError(msg)

        masks = []
        for bbox, edges in zip(self._bbox, self._centered_edges, strict=True):
            ny, nx = bbox.shape
            theta_rad = self.theta.to(u.radian).value
            mask = elliptical_overlap_grid(edges[0], edges[1], edges[2],
                                           edges[3], nx, ny, a, b,
                                           theta_rad, use_exact, subpixels)

            # Subtract the inner ellipse for an annulus
            if hasattr(self, 'a_in'):
                mask -= elliptical_overlap_grid(edges[0], edges[1], edges[2],
                                                edges[3], nx, ny, self.a_in,
                                                self.b_in, theta_rad,
                                                use_exact, subpixels)

            masks.append(ApertureMask(mask, bbox))

        if self.isscalar:
            return masks[0]

        return masks

    @staticmethod
    def _calc_extents(semimajor_axis, semiminor_axis, theta):
        """
        Calculate half of the bounding box extents of an ellipse.
        """
        return _calc_ellipse_extents(semimajor_axis, semiminor_axis, theta)


def _calc_ellipse_extents(semimajor_axis, semiminor_axis, theta):
    """
    Calculate half of the bounding box extents of an ellipse.
    """
    theta_rad = theta.to(u.radian).value
    cos_theta = np.cos(theta_rad)
    sin_theta = np.sin(theta_rad)
    semimajor_x = semimajor_axis * cos_theta
    semimajor_y = semimajor_axis * sin_theta
    semiminor_x = semiminor_axis * -sin_theta
    semiminor_y = semiminor_axis * cos_theta
    x_extent = np.sqrt(semimajor_x**2 + semiminor_x**2)
    y_extent = np.sqrt(semimajor_y**2 + semiminor_y**2)

    return x_extent, y_extent


class EllipticalAperture(PixelAperture):
    """
    An elliptical aperture defined in pixel coordinates.

    The aperture has a single fixed size/shape, but it can have multiple
    positions (see the ``positions`` input).

    Parameters
    ----------
    positions : array_like
        The pixel coordinates of the aperture center(s) in one of the
        following formats:

        * single ``(x, y)`` pair as a tuple, list, or `~numpy.ndarray`
        * tuple, list, or `~numpy.ndarray` of ``(x, y)`` pairs

    a : float
        The semimajor axis of the ellipse in pixels.

    b : float
        The semiminor axis of the ellipse in pixels.

    theta : float or `~astropy.units.Quantity`, optional
        The rotation angle as an angular quantity
        (`~astropy.units.Quantity` or `~astropy.coordinates.Angle`) or
        value in radians (as a float) from the positive ``x`` axis. The
        rotation angle increases counterclockwise.

    Raises
    ------
    ValueError : `ValueError`
        If either axis (``a`` or ``b``) is negative.

    Examples
    --------
    >>> from astropy.coordinates import Angle
    >>> from photutils.aperture import EllipticalAperture

    >>> theta = Angle(80, 'deg')
    >>> aper = EllipticalAperture([10.0, 20.0], 5.0, 3.0)
    >>> aper = EllipticalAperture((10.0, 20.0), 5.0, 3.0, theta=theta)

    >>> pos1 = (10.0, 20.0)  # (x, y)
    >>> pos2 = (30.0, 40.0)
    >>> pos3 = (50.0, 60.0)
    >>> aper = EllipticalAperture([pos1, pos2, pos3], 5.0, 3.0)
    >>> aper = EllipticalAperture((pos1, pos2, pos3), 5.0, 3.0, theta=theta)
    """

    _params = ('positions', 'a', 'b', 'theta')
    positions = PixelPositions('The center pixel position(s).')
    a = PositiveScalar('The semimajor axis in pixels.')
    b = PositiveScalar('The semiminor axis in pixels.')
    theta = ScalarAngleOrValue('The counterclockwise rotation angle as an '
                               'angular Quantity or value in radians from '
                               'the positive x axis.')

    @deprecated_positional_kwargs(since='3.0', until='4.0')
    def __init__(self, positions, a, b, theta=0.0):
        self.positions = positions
        self.a = a
        self.b = b
        self.theta = theta

    @lazyproperty
    def _xy_extents(self):
        """
        The half of the bounding box extents of the ellipse in the x and
        y directions.
        """
        return _calc_ellipse_extents(self.a, self.b, self.theta)

    @lazyproperty
    def area(self):
        """
        The exact geometric area of the aperture shape.
        """
        return math.pi * self.a * self.b

    def _to_patch(self, *, origin=(0, 0), **kwargs):
        """
        Return a `~matplotlib.patches.Patch` for the aperture.

        Parameters
        ----------
        origin : array_like, optional
            The ``(x, y)`` position of the origin of the displayed
            image.

        **kwargs : dict, optional
            Any keyword arguments accepted by
            `matplotlib.patches.Patch`.

        Returns
        -------
        patch : `~matplotlib.patches.Patch` or list of \
                `~matplotlib.patches.Patch`
            A patch for the aperture. If the aperture is scalar then a
            single `~matplotlib.patches.Patch` is returned, otherwise a
            list of `~matplotlib.patches.Patch` is returned.
        """
        import matplotlib.patches as mpatches

        xy_positions, patch_kwargs = self._define_patch_params(origin=origin,
                                                               **kwargs)

        angle = self.theta.to(u.deg).value
        patches = [mpatches.Ellipse(xy_position, 2.0 * self.a, 2.0 * self.b,
                                    angle=angle, **patch_kwargs)
                   for xy_position in xy_positions]

        if self.isscalar:
            return patches[0]

        return patches

    def _compute_overlap(self, edges, nx, ny, use_exact, subpixels):
        """
        Compute the overlap of the aperture on the pixel grid.

        Parameters
        ----------
        edges : list of 4 1D `~numpy.ndarray`
            The edges of the pixel grid in the form of
            ``[x_edges, y_edges, x_centers, y_centers]``.

        nx, ny : int
            The number of pixels in the x and y directions.

        use_exact : bool
            Whether to use the exact method for calculating the overlap.

        subpixels : int
            The number of subpixels to use in each dimension for the
            subpixel method.

        Returns
        -------
        overlap : 2D `~numpy.ndarray`
            The overlap of the aperture on the pixel grid. The values
            will be between 0 and 1, where 0 means no overlap and 1
            means full overlap.
        """
        theta_rad = self.theta.to(u.radian).value
        return elliptical_overlap_grid(edges[0], edges[1], edges[2],
                                       edges[3], nx, ny, self.a, self.b,
                                       theta_rad, use_exact, subpixels)

    def to_sky(self, wcs):
        """
        Convert the aperture to a `SkyEllipticalAperture` object defined
        in celestial coordinates.

        Parameters
        ----------
        wcs : WCS object
            A world coordinate system (WCS) transformation that
            supports the `astropy shared interface for WCS
            <https://docs.astropy.org/en/stable/wcs/wcsapi.html>`_
            (e.g., `astropy.wcs.WCS`, `gwcs.wcs.WCS`).

        Returns
        -------
        aperture : `SkyEllipticalAperture` object
            A `SkyEllipticalAperture` object.

        Notes
        -----
        The aperture shape parameters are converted using the local WCS
        properties (pixel scale, rotation angle) evaluated at the first
        aperture position. Because aperture objects require scalar shape
        parameters, only a single reference position is used for the
        conversion. For apertures with multiple positions used with a
        WCS that has spatially-varying distortions, this may produce
        inaccurate results for positions far from the first position.
        """
        xpos, ypos = np.transpose(self.positions)
        positions = wcs.pixel_to_world(xpos, ypos)

        first_pos = np.atleast_2d(self.positions)[0]
        pixcoord = (float(first_pos[0]), float(first_pos[1]))
        _, sky_width, sky_height, sky_angle = pixel_ellipse_to_sky_svd(
            pixcoord, wcs, 2 * self.a, 2 * self.b, self.theta.to(u.rad).value)

        a = Angle(sky_width / 2, 'arcsec')
        b = Angle(sky_height / 2, 'arcsec')
        return SkyEllipticalAperture(positions=positions, a=a, b=b,
                                     theta=sky_angle)


class EllipticalAnnulus(PixelAperture):
    r"""
    An elliptical annulus aperture defined in pixel coordinates.

    The aperture has a single fixed size/shape, but it can have multiple
    positions (see the ``positions`` input).

    Parameters
    ----------
    positions : array_like
        The pixel coordinates of the aperture center(s) in one of the
        following formats:

        * single ``(x, y)`` pair as a tuple, list, or `~numpy.ndarray`
        * tuple, list, or `~numpy.ndarray` of ``(x, y)`` pairs

    a_in : float
        The inner semimajor axis of the elliptical annulus in pixels.

    a_out : float
        The outer semimajor axis of the elliptical annulus in pixels.

    b_out : float
        The outer semiminor axis of the elliptical annulus in pixels.

    b_in : `None` or float, optional
        The inner semiminor axis of the elliptical annulus in pixels.
        If `None`, then the inner semiminor axis is calculated as:

        .. math::

            b_{in} = b_{out} \left(\frac{a_{in}}{a_{out}}\right)

    theta : float or `~astropy.units.Quantity`, optional
        The rotation angle as an angular quantity
        (`~astropy.units.Quantity` or `~astropy.coordinates.Angle`) or
        value in radians (as a float) from the positive ``x`` axis. The
        rotation angle increases counterclockwise.

    Raises
    ------
    ValueError : `ValueError`
        If inner semimajor axis (``a_in``) is greater than outer semimajor
        axis (``a_out``).

    ValueError : `ValueError`
        If either the inner semimajor axis (``a_in``) or the outer semiminor
        axis (``b_out``) is negative.

    Examples
    --------
    >>> from astropy.coordinates import Angle
    >>> from photutils.aperture import EllipticalAnnulus

    >>> theta = Angle(80, 'deg')
    >>> aper = EllipticalAnnulus([10.0, 20.0], 3.0, 8.0, 5.0)
    >>> aper = EllipticalAnnulus((10.0, 20.0), 3.0, 8.0, 5.0, theta=theta)

    >>> pos1 = (10.0, 20.0)  # (x, y)
    >>> pos2 = (30.0, 40.0)
    >>> pos3 = (50.0, 60.0)
    >>> aper = EllipticalAnnulus([pos1, pos2, pos3], 3.0, 8.0, 5.0)
    >>> aper = EllipticalAnnulus((pos1, pos2, pos3), 3.0, 8.0, 5.0,
    ...                          theta=theta)
    """

    _params = ('positions', 'a_in', 'a_out', 'b_in', 'b_out', 'theta')
    positions = PixelPositions('The center pixel position(s).')
    a_in = PositiveScalar('The inner semimajor axis in pixels.')
    a_out = PositiveScalar('The outer semimajor axis in pixels.')
    b_in = PositiveScalar('The inner semiminor axis in pixels.')
    b_out = PositiveScalar('The outer semiminor axis in pixels.')
    theta = ScalarAngleOrValue('The counterclockwise rotation angle as an '
                               'angular Quantity or value in radians from '
                               'the positive x axis.')

    @deprecated_positional_kwargs(since='3.0', until='4.0')
    def __init__(self, positions, a_in, a_out, b_out, b_in=None, theta=0.0):
        if not a_out > a_in:
            msg = "'a_out' must be greater than 'a_in'"
            raise ValueError(msg)

        self.positions = positions
        self.a_in = a_in
        self.a_out = a_out
        self.b_out = b_out

        if b_in is None:
            b_in = self.b_out * self.a_in / self.a_out
        elif not b_out > b_in:
            msg = "'b_out' must be greater than 'b_in'"
            raise ValueError(msg)
        self.b_in = b_in

        self.theta = theta

    @lazyproperty
    def _xy_extents(self):
        """
        The half of the bounding box extents of the outer ellipse in the
        x and y directions.
        """
        return _calc_ellipse_extents(self.a_out, self.b_out, self.theta)

    @lazyproperty
    def area(self):
        """
        The exact geometric area of the aperture shape.
        """
        return math.pi * (self.a_out * self.b_out - self.a_in * self.b_in)

    def _to_patch(self, *, origin=(0, 0), **kwargs):
        """
        Return a `~matplotlib.patches.Patch` for the aperture.

        Parameters
        ----------
        origin : array_like, optional
            The ``(x, y)`` position of the origin of the displayed
            image.

        **kwargs : dict, optional
            Any keyword arguments accepted by
            `matplotlib.patches.Patch`.

        Returns
        -------
        patch : `~matplotlib.patches.Patch` or list of \
                `~matplotlib.patches.Patch`
            A patch for the aperture. If the aperture is scalar then a
            single `~matplotlib.patches.Patch` is returned, otherwise a
            list of `~matplotlib.patches.Patch` is returned.
        """
        import matplotlib.patches as mpatches

        xy_positions, patch_kwargs = self._define_patch_params(origin=origin,
                                                               **kwargs)

        patches = []
        angle = self.theta.to(u.deg).value
        for xy_position in xy_positions:
            patch_inner = mpatches.Ellipse(xy_position, 2.0 * self.a_in,
                                           2.0 * self.b_in, angle=angle)
            patch_outer = mpatches.Ellipse(xy_position, 2.0 * self.a_out,
                                           2.0 * self.b_out, angle=angle)
            path = self._make_annulus_path(patch_inner, patch_outer)
            patches.append(mpatches.PathPatch(path, **patch_kwargs))

        if self.isscalar:
            return patches[0]

        return patches

    def _compute_overlap(self, edges, nx, ny, use_exact, subpixels):
        """
        Compute the overlap of the aperture on the pixel grid.

        Parameters
        ----------
        edges : list of 4 1D `~numpy.ndarray`
            The edges of the pixel grid in the form of
            ``[x_edges, y_edges, x_centers, y_centers]``.

        nx, ny : int
            The number of pixels in the x and y directions.

        use_exact : bool
            Whether to use the exact method for calculating the overlap.

        subpixels : int
            The number of subpixels to use in each dimension for the
            subpixel method.

        Returns
        -------
        overlap : 2D `~numpy.ndarray`
            The overlap of the aperture on the pixel grid. The values
            will be between 0 and 1, where 0 means no overlap and 1
            means full overlap.
        """
        theta_rad = self.theta.to(u.radian).value
        overlap = elliptical_overlap_grid(edges[0], edges[1], edges[2],
                                          edges[3], nx, ny, self.a_out,
                                          self.b_out, theta_rad,
                                          use_exact, subpixels)
        overlap -= elliptical_overlap_grid(edges[0], edges[1], edges[2],
                                           edges[3], nx, ny, self.a_in,
                                           self.b_in, theta_rad,
                                           use_exact, subpixels)
        return overlap

    def to_sky(self, wcs):
        """
        Convert the aperture to a `SkyEllipticalAnnulus` object defined
        in celestial coordinates.

        Parameters
        ----------
        wcs : WCS object
            A world coordinate system (WCS) transformation that
            supports the `astropy shared interface for WCS
            <https://docs.astropy.org/en/stable/wcs/wcsapi.html>`_
            (e.g., `astropy.wcs.WCS`, `gwcs.wcs.WCS`).

        Returns
        -------
        aperture : `SkyEllipticalAnnulus` object
            A `SkyEllipticalAnnulus` object.

        Notes
        -----
        The aperture shape parameters are converted using the local WCS
        properties (pixel scale, rotation angle) evaluated at the first
        aperture position. Because aperture objects require scalar shape
        parameters, only a single reference position is used for the
        conversion. For apertures with multiple positions used with a
        WCS that has spatially-varying distortions, this may produce
        inaccurate results for positions far from the first position.
        """
        xpos, ypos = np.transpose(self.positions)
        positions = wcs.pixel_to_world(xpos, ypos)

        first_pos = np.atleast_2d(self.positions)[0]
        pixcoord = (float(first_pos[0]), float(first_pos[1]))
        theta_rad = self.theta.to(u.rad).value

        _, sky_w_out, sky_h_out, sky_angle = pixel_ellipse_to_sky_svd(
            pixcoord, wcs, 2 * self.a_out, 2 * self.b_out, theta_rad)
        _, sky_w_in, sky_h_in, _ = pixel_ellipse_to_sky_svd(
            pixcoord, wcs, 2 * self.a_in, 2 * self.b_in, theta_rad)

        a_out = Angle(sky_w_out / 2, 'arcsec')
        b_out = Angle(sky_h_out / 2, 'arcsec')
        a_in = Angle(sky_w_in / 2, 'arcsec')
        b_in = Angle(sky_h_in / 2, 'arcsec')
        return SkyEllipticalAnnulus(positions=positions, a_in=a_in,
                                    a_out=a_out, b_out=b_out,
                                    b_in=b_in, theta=sky_angle)


class SkyEllipticalAperture(SkyAperture):
    """
    An elliptical aperture defined in sky coordinates.

    The aperture has a single fixed size/shape, but it can have multiple
    positions (see the ``positions`` input).

    Parameters
    ----------
    positions : `~astropy.coordinates.SkyCoord`
        The celestial coordinates of the aperture center(s). This can be
        either scalar coordinates or an array of coordinates.

    a : scalar `~astropy.units.Quantity`
        The semimajor axis of the ellipse in angular units.

    b : scalar `~astropy.units.Quantity`
        The semiminor axis of the ellipse in angular units.

    theta : scalar `~astropy.units.Quantity`, optional
        The position angle (in angular units) of the ellipse semimajor
        axis. For a right-handed world coordinate system, the position
        angle increases counterclockwise from North (PA=0).

    Examples
    --------
    >>> from astropy.coordinates import SkyCoord
    >>> import astropy.units as u
    >>> from photutils.aperture import SkyEllipticalAperture
    >>> positions = SkyCoord(ra=[10.0, 20.0], dec=[30.0, 40.0], unit='deg')
    >>> aper = SkyEllipticalAperture(positions, 1.0*u.arcsec, 0.5*u.arcsec)
    """

    _params = ('positions', 'a', 'b', 'theta')
    positions = SkyCoordPositions('The center position(s) in sky coordinates.')
    a = PositiveScalarAngle('The semimajor axis in angular units.')
    b = PositiveScalarAngle('The semiminor axis in angular units.')
    theta = ScalarAngle('The position angle in angular units of the ellipse '
                        'semimajor axis.')

    @deprecated_positional_kwargs(since='3.0', until='4.0')
    def __init__(self, positions, a, b, theta=0.0 * u.deg):
        self.positions = positions
        self.a = a
        self.b = b
        self.theta = theta

    def to_pixel(self, wcs):
        """
        Convert the aperture to an `EllipticalAperture` object defined
        in pixel coordinates.

        Parameters
        ----------
        wcs : WCS object
            A world coordinate system (WCS) transformation that
            supports the `astropy shared interface for WCS
            <https://docs.astropy.org/en/stable/wcs/wcsapi.html>`_
            (e.g., `astropy.wcs.WCS`, `gwcs.wcs.WCS`).

        Returns
        -------
        aperture : `EllipticalAperture` object
            An `EllipticalAperture` object.

        Notes
        -----
        The aperture shape parameters are converted using the local WCS
        properties (pixel scale, rotation angle) evaluated at the first
        aperture position. Because aperture objects require scalar shape
        parameters, only a single reference position is used for the
        conversion. For apertures with multiple positions used with a
        WCS that has spatially-varying distortions, this may produce
        inaccurate results for positions far from the first position.
        """
        xpos, ypos = wcs.world_to_pixel(self.positions)
        positions = np.transpose((xpos, ypos))

        skypos = self.positions if self.isscalar else self.positions[0]
        sky_angle_rad = self.theta.to(u.rad).value
        _, pix_width, pix_height, pix_angle = sky_ellipse_to_pixel_svd(
            skypos, wcs,
            2 * self.a.to(u.arcsec).value,
            2 * self.b.to(u.arcsec).value,
            sky_angle_rad)

        a = pix_width / 2
        b = pix_height / 2
        return EllipticalAperture(positions=positions, a=a, b=b,
                                  theta=pix_angle)


class SkyEllipticalAnnulus(SkyAperture):
    r"""
    An elliptical annulus aperture defined in sky coordinates.

    The aperture has a single fixed size/shape, but it can have multiple
    positions (see the ``positions`` input).

    Parameters
    ----------
    positions : `~astropy.coordinates.SkyCoord`
        The celestial coordinates of the aperture center(s). This can be
        either scalar coordinates or an array of coordinates.

    a_in : scalar `~astropy.units.Quantity`
        The inner semimajor axis in angular units.

    a_out : scalar `~astropy.units.Quantity`
        The outer semimajor axis in angular units.

    b_out : scalar `~astropy.units.Quantity`
        The outer semiminor axis in angular units.

    b_in : `None` or scalar `~astropy.units.Quantity`
        The inner semiminor axis in angular units. If `None`, then the
        inner semiminor axis is calculated as:

        .. math::

            b_{in} = b_{out} \left(\frac{a_{in}}{a_{out}}\right)

    theta : scalar `~astropy.units.Quantity`, optional
        The position angle (in angular units) of the ellipse semimajor
        axis. For a right-handed world coordinate system, the position
        angle increases counterclockwise from North (PA=0).

    Examples
    --------
    >>> from astropy.coordinates import SkyCoord
    >>> import astropy.units as u
    >>> from photutils.aperture import SkyEllipticalAnnulus
    >>> positions = SkyCoord(ra=[10.0, 20.0], dec=[30.0, 40.0], unit='deg')
    >>> aper = SkyEllipticalAnnulus(positions, 0.5*u.arcsec, 2.0*u.arcsec,
    ...                             1.0*u.arcsec)
    """

    _params = ('positions', 'a_in', 'a_out', 'b_in', 'b_out', 'theta')
    positions = SkyCoordPositions('The center position(s) in sky coordinates.')
    a_in = PositiveScalarAngle('The inner semimajor axis in angular units.')
    a_out = PositiveScalarAngle('The outer semimajor axis in angular units.')
    b_in = PositiveScalarAngle('The inner semiminor axis in angular units.')
    b_out = PositiveScalarAngle('The outer semiminor axis in angular units.')
    theta = ScalarAngle('The position angle in angular units of the ellipse '
                        'semimajor axis.')

    @deprecated_positional_kwargs(since='3.0', until='4.0')
    def __init__(self, positions, a_in, a_out, b_out, b_in=None,
                 theta=0.0 * u.deg):
        if not a_out > a_in:
            msg = "'a_out' must be greater than 'a_in'"
            raise ValueError(msg)

        self.positions = positions
        self.a_in = a_in
        self.a_out = a_out
        self.b_out = b_out

        if b_in is None:
            b_in = self.b_out * self.a_in / self.a_out
        elif not b_out > b_in:
            msg = "'b_out' must be greater than 'b_in'"
            raise ValueError(msg)
        self.b_in = b_in

        self.theta = theta

    def to_pixel(self, wcs):
        """
        Convert the aperture to an `EllipticalAnnulus` object defined in
        pixel coordinates.

        Parameters
        ----------
        wcs : WCS object
            A world coordinate system (WCS) transformation that
            supports the `astropy shared interface for WCS
            <https://docs.astropy.org/en/stable/wcs/wcsapi.html>`_
            (e.g., `astropy.wcs.WCS`, `gwcs.wcs.WCS`).

        Returns
        -------
        aperture : `EllipticalAnnulus` object
            An `EllipticalAnnulus` object.

        Notes
        -----
        The aperture shape parameters are converted using the local WCS
        properties (pixel scale, rotation angle) evaluated at the first
        aperture position. Because aperture objects require scalar shape
        parameters, only a single reference position is used for the
        conversion. For apertures with multiple positions used with a
        WCS that has spatially-varying distortions, this may produce
        inaccurate results for positions far from the first position.
        """
        xpos, ypos = wcs.world_to_pixel(self.positions)
        positions = np.transpose((xpos, ypos))

        skypos = self.positions if self.isscalar else self.positions[0]
        sky_angle_rad = self.theta.to(u.rad).value

        _, pix_w_out, pix_h_out, pix_angle = sky_ellipse_to_pixel_svd(
            skypos, wcs,
            2 * self.a_out.to(u.arcsec).value,
            2 * self.b_out.to(u.arcsec).value,
            sky_angle_rad)
        _, pix_w_in, pix_h_in, _ = sky_ellipse_to_pixel_svd(
            skypos, wcs,
            2 * self.a_in.to(u.arcsec).value,
            2 * self.b_in.to(u.arcsec).value,
            sky_angle_rad)

        a_out = pix_w_out / 2
        b_out = pix_h_out / 2
        a_in = pix_w_in / 2
        b_in = pix_h_in / 2
        return EllipticalAnnulus(positions=positions, a_in=a_in,
                                 a_out=a_out, b_out=b_out,
                                 b_in=b_in, theta=pix_angle)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tools for generating random (x, y) coordinates.
"""

import warnings

import numpy as np
from astropy.utils.exceptions import AstropyUserWarning
from scipy.spatial import KDTree


def apply_separation(xycoords, min_separation):
    """
    Apply a minimum separation to a set of (x, y) coordinates.

    Coordinates that are closer than the minimum separation are removed.

    Parameters
    ----------
    xycoords : `~numpy.ndarray`
        The (x, y) coordinates with shape ``(N, 2)``.

    min_separation : float
        The minimum separation in pixels between coordinates.

    Returns
    -------
    xycoords : `~numpy.ndarray`
        The (x, y) coordinates with shape ``(N, 2)`` after excluding
        points closer than the minimum separation.
    """
    tree = KDTree(xycoords)
    pairs = tree.query_pairs(min_separation, output_type='ndarray')

    if len(pairs) == 0:
        return xycoords

    n = xycoords.shape[0]
    keep = np.ones(n, dtype=bool)

    # Group pairs by first index for vectorized neighbor removal. Each
    # pair has i < j (guaranteed by KDTree). Process groups in ascending
    # first-index order (greedy independent set algorithm): for each
    # kept point, discard all its higher-index neighbors.
    sorted_idx = pairs[:, 0].argsort(kind='stable')
    pairs_sorted = pairs[sorted_idx]
    unique_i, group_start = np.unique(pairs_sorted[:, 0],
                                      return_index=True)
    group_end = np.append(group_start[1:], len(pairs_sorted))

    for k in range(len(unique_i)):
        i = unique_i[k]
        if keep[i]:
            js = pairs_sorted[group_start[k]:group_end[k], 1]
            keep[js] = False

    return xycoords[keep]


def make_random_xycoords(size, x_range, y_range, *, min_separation=0.0,
                         seed=None, oversample=10):
    """
    Make random (x, y) coordinates.

    Parameters
    ----------
    size : int
        The number of coordinates to generate.

    x_range : tuple
        The range of x values (min, max).

    y_range : tuple
        The range of y values (min, max).

    min_separation : float, optional
        The minimum separation in pixels between coordinates.

    seed : int, optional
        A seed to initialize the `numpy.random.BitGenerator`. If `None`,
        then fresh, unpredictable entropy will be pulled from the OS.

    oversample : int, optional
        The oversampling factor used when ``min_separation`` > 0 to
        generate extra candidate coordinates before filtering by
        separation. Higher values increase the chance of producing the
        requested number of coordinates in crowded conditions, at the
        cost of speed. The default is 10.

    Returns
    -------
    xycoords : `~numpy.ndarray`
        The (x, y) random coordinates with shape ``(size, 2)``. When
        ``min_separation`` > 0, fewer than ``size`` coordinates may be
        returned if the range and separation cannot be satisfied. A
        warning is issued in that case.
    """
    if size == 0:
        return np.empty((0, 2))

    if x_range[0] >= x_range[1] or y_range[0] >= y_range[1]:
        msg = 'x_range and y_range must be (min, max) with min < max.'
        raise ValueError(msg)

    ncoords = size
    if min_separation > 0:
        # Scale the number of random coordinates to account for
        # some being discarded due to min_separation
        ncoords *= oversample

    rng = np.random.default_rng(seed)
    xc = rng.uniform(x_range[0], x_range[1], ncoords)
    yc = rng.uniform(y_range[0], y_range[1], ncoords)
    xycoords = np.transpose(np.array((xc, yc)))

    xycoords = apply_separation(xycoords, min_separation)
    xycoords = xycoords[:size]

    if len(xycoords) < size:
        msg = (f'Unable to produce {size!r} coordinates within the '
               'given shape and minimum separation. Only '
               f'{len(xycoords)!r} coordinates were generated.')
        warnings.warn(msg, AstropyUserWarning)

    return xycoords



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the utils module.
"""

import numpy as np
import pytest
from numpy.testing import assert_allclose
from scipy.fft import fft2

from photutils.psf_matching.tests.conftest import _make_gaussian_psf
from photutils.psf_matching.utils import (_apply_window_to_fourier,
                                          _convert_psf_to_otf, _validate_psf,
                                          _validate_window_array, resize_psf)


class TestValidatePSF:
    def test_valid_psf(self):
        """
        Test that a valid PSF passes validation without error.
        """
        psf = _make_gaussian_psf(5, 1.5)
        _validate_psf(psf, 'psf')  # should not raise

    def test_non_2d(self):
        """
        Test that non-2D array raises ValueError.
        """
        match = 'psf must be a 2D array'
        with pytest.raises(ValueError, match=match):
            _validate_psf(np.ones(5), 'psf')

    def test_even_shape(self):
        """
        Test that even-shaped array raises ValueError.
        """
        psf = np.zeros((4, 4))
        psf[2, 2] = 1.0
        match = 'must have odd dimensions'
        with pytest.raises(ValueError, match=match):
            _validate_psf(psf, 'psf')

    def test_nan_inf_values(self):
        """
        Test that NaN or Inf values raise ValueError.
        """
        psf = np.zeros((25, 25))
        psf[12, 12] = np.nan
        match = 'contains NaN or Inf values'
        with pytest.raises(ValueError, match=match):
            _validate_psf(psf, 'psf')

        psf[12, 12] = np.inf
        with pytest.raises(ValueError, match=match):
            _validate_psf(psf, 'psf')

    def test_zero_psf_raises(self):
        """
        Test that an all-zero PSF raises ValueError (cannot be
        normalized).
        """
        psf = np.zeros((5, 5))
        match = 'must have a non-zero sum'
        with pytest.raises(ValueError, match=match):
            _validate_psf(psf, 'psf')


class TestValidateWindowArray:
    def test_valid_window(self):
        """
        Test that a valid window array passes validation.
        """
        shape = (25, 25)
        window = np.ones(shape)
        _validate_window_array(window, shape)  # should not raise

    def test_not_2d(self):
        """
        Test that non-2D window raises ValueError.
        """
        match = 'window function must return a 2D array'
        with pytest.raises(ValueError, match=match):
            _validate_window_array(np.ones(10), (10,))

    def test_wrong_shape(self):
        """
        Test that wrong-shaped window raises ValueError.
        """
        match = 'window function must return an array with shape'
        with pytest.raises(ValueError, match=match):
            _validate_window_array(np.ones((10, 10)), (25, 25))

    def test_values_below_zero(self):
        """
        Test that window values < 0 raise ValueError.
        """
        arr = np.ones((10, 10))
        arr[0, 0] = -0.1
        match = 'window function values must be in the range'
        with pytest.raises(ValueError, match=match):
            _validate_window_array(arr, (10, 10))

    def test_values_above_one(self):
        """
        Test that window values > 1 raise ValueError.
        """
        arr = np.ones((10, 10))
        arr[0, 0] = 1.5
        match = 'window function values must be in the range'
        with pytest.raises(ValueError, match=match):
            _validate_window_array(arr, (10, 10))


class TestConvertPsfToOtf:
    def test_zero_psf(self):
        """
        Test that an all-zero PSF returns an all-zero OTF.
        """
        psf = np.zeros((3, 3))
        otf = _convert_psf_to_otf(psf, (5, 5))
        assert_allclose(otf, 0.0)

    def test_invalid_psf(self):
        match = 'psf must be a 2D array'
        with pytest.raises(ValueError, match=match):
            _convert_psf_to_otf(np.ones(5), (5, 5))

        match = 'psf must have odd dimensions'
        with pytest.raises(ValueError, match=match):
            _convert_psf_to_otf(np.ones((6, 6)), (11, 11))

    def test_output_shape(self):
        """
        Test that the output OTF has the requested shape.
        """
        psf = np.array([[0, -1, 0],
                        [-1, 4, -1],
                        [0, -1, 0]])
        otf = _convert_psf_to_otf(psf, (25, 25))
        assert otf.shape == (25, 25)

    def test_delta_function(self):
        """
        Test that a delta function produces a flat OTF.
        """
        # A single-pixel PSF at the center of a 1x1 array
        psf = np.array([[1.0]])
        otf = _convert_psf_to_otf(psf, (5, 5))
        assert_allclose(np.abs(otf), 1.0)

    def test_power_spectrum_shift_invariant(self):
        """
        Test that |OTF|^2 is the same regardless of PSF centering.

        The power spectrum should be independent of the input kernel
        position because the circular shift only affects the phase.
        """
        laplacian = np.array([[0, -1, 0],
                              [-1, 4, -1],
                              [0, -1, 0]])
        shape = (25, 25)

        otf = _convert_psf_to_otf(laplacian, shape)
        power = np.abs(otf) ** 2

        # Compare to a naive approach (no circular shift)
        padded = np.zeros(shape)
        padded[:3, :3] = laplacian
        otf_naive = fft2(padded)
        power_naive = np.abs(otf_naive) ** 2

        assert_allclose(power, power_naive)

    def test_psf_larger_than_shape(self):
        """
        Test that a PSF larger than the target shape raises ValueError.
        """
        psf = np.ones((7, 7))
        match = 'PSF shape.*is larger than the target shape'
        with pytest.raises(ValueError, match=match):
            _convert_psf_to_otf(psf, (5, 5))

    def test_laplacian_dc_is_zero(self):
        """
        Test that the Laplacian OTF has zero power at DC.

        The Laplacian sums to zero, so its DC component should be zero.
        """
        laplacian = np.array([[0, -1, 0],
                              [-1, 4, -1],
                              [0, -1, 0]])
        otf = _convert_psf_to_otf(laplacian, (25, 25))
        assert_allclose(otf[0, 0], 0.0, atol=1e-15)

    def test_normalized_psf_dc_component(self):
        """
        Test that a normalized PSF has DC component equal to 1.0.

        For a normalized PSF (sum = 1), the DC component of the OTF
        should be 1.0.
        """
        psf = _make_gaussian_psf(5, 1.5)
        otf = _convert_psf_to_otf(psf, (25, 25))
        assert_allclose(otf[0, 0], 1.0, rtol=1e-10)

    def test_non_square_shapes(self):
        """
        Test with non-square PSF and output shapes.
        """
        # 3x5 PSF to 11x21 output
        psf = np.zeros((3, 5))
        psf[1, 2] = 1.0  # center at (1, 2)
        otf = _convert_psf_to_otf(psf, (11, 21))
        assert otf.shape == (11, 21)
        assert_allclose(np.abs(otf), 1.0, rtol=1e-10)

    def test_no_padding_needed(self):
        """
        Test when PSF size equals output size (no padding).
        """
        psf = _make_gaussian_psf(7, 1.5)
        otf = _convert_psf_to_otf(psf, (7, 7))
        assert otf.shape == (7, 7)
        assert_allclose(otf[0, 0], 1.0, rtol=1e-10)

    def test_symmetric_psf_real_otf(self):
        """
        Test that a symmetric PSF produces a nearly real OTF.

        A centered, symmetric PSF should produce an OTF with negligible
        imaginary components.
        """
        psf = _make_gaussian_psf(7, 2.0)
        otf = _convert_psf_to_otf(psf, (25, 25))
        # Imaginary components should be negligible
        assert np.max(np.abs(otf.imag)) < 1e-10

    def test_larger_psf_sizes(self):
        """
        Test with various PSF sizes to ensure centering works correctly.
        """
        for psf_size in [5, 7, 9]:
            psf = _make_gaussian_psf(psf_size, psf_size / 4.0)
            otf = _convert_psf_to_otf(psf, (25, 25))
            # Normalized PSF should have DC component of 1.0
            assert_allclose(otf[0, 0], 1.0, rtol=1e-10)
            # OTF should be nearly real for symmetric PSF
            assert np.max(np.abs(otf.imag)) < 1e-10


class TestResizePSF:
    def test_resize(self):
        """
        Test that resizing returns an odd-shaped output.

        For a (5,5) input with ratio=2.0, ceil gives 10 (even), so one
        pixel is added to give (11, 11).
        """
        psf = _make_gaussian_psf(5, 1.5)
        result = resize_psf(psf, 0.1, 0.05)
        assert result.shape == (11, 11)
        assert_allclose(result.sum(), psf.sum())

    def test_resize_odd_output(self):
        """
        Test that resizing to a naturally odd output shape is unchanged.
        """
        psf = _make_gaussian_psf(5, 1.5)
        result = resize_psf(psf, 0.1, 0.1)  # ratio=1.0 -> 5x5 (odd)
        assert result.shape == (5, 5)

    def test_resize_always_odd(self):
        """
        Test that the output is always odd across a range of ratios.
        """
        psf = _make_gaussian_psf(5, 1.5)
        for scale_out in [0.04, 0.05, 0.06, 0.07, 0.08]:
            result = resize_psf(psf, 0.1, scale_out)
            assert result.shape[0] % 2 == 1
            assert result.shape[1] % 2 == 1

    def test_non_2d(self):
        """
        Test that non-2D PSF raises ValueError.
        """
        match = 'psf must be a 2D array'
        with pytest.raises(ValueError, match=match):
            resize_psf(np.ones(5), 0.1, 0.05)

    def test_even_shape(self):
        """
        Test that even-shaped PSF raises ValueError.
        """
        psf = np.zeros((4, 4))
        psf[2, 2] = 1.0
        match = 'must have odd dimensions'
        with pytest.raises(ValueError, match=match):
            resize_psf(psf, 0.1, 0.05)

    def test_non_positive_input_scale(self):
        """
        Test that negative input_pixel_scale raises ValueError.
        """
        psf = _make_gaussian_psf(5, 1.5)
        match = 'must be positive'
        with pytest.raises(ValueError, match=match):
            resize_psf(psf, -0.1, 0.05)

    def test_non_positive_output_scale(self):
        """
        Test that negative output_pixel_scale raises ValueError.
        """
        psf = _make_gaussian_psf(5, 1.5)
        match = 'must be positive'
        with pytest.raises(ValueError, match=match):
            resize_psf(psf, 0.1, -0.05)

    def test_zero_scale(self):
        """
        Test that zero input_pixel_scale raises ValueError.
        """
        psf = _make_gaussian_psf(5, 1.5)
        match = 'must be positive'
        with pytest.raises(ValueError, match=match):
            resize_psf(psf, 0.0, 0.05)


class TestApplyWindowToFourier:
    def test_basic(self):
        """
        Test that _apply_window_to_fourier applies a window to a
        Fourier array.
        """
        shape = (11, 11)
        fourier_array = np.ones(shape, dtype=complex)

        def uniform_window(shape):
            return np.ones(shape)

        result = _apply_window_to_fourier(fourier_array, uniform_window,
                                          shape)
        assert result.shape == shape
        assert np.allclose(result, fourier_array)

    def test_zero_window(self):
        """
        Test that a zero window zeros out the Fourier array.
        """
        shape = (11, 11)
        fourier_array = np.ones(shape, dtype=complex)

        def zero_window(shape):
            return np.zeros(shape)

        result = _apply_window_to_fourier(fourier_array, zero_window, shape)
        assert np.allclose(result, 0.0)

    def test_invalid_window_raises(self):
        """
        Test that an invalid window function raises ValueError.
        """
        shape = (11, 11)
        fourier_array = np.ones(shape, dtype=complex)

        def bad_window(shape):  # noqa: ARG001
            return np.ones((5, 5))  # wrong shape

        match = 'window function must return an array with shape'
        with pytest.raises(ValueError, match=match):
            _apply_window_to_fourier(fourier_array, bad_window, shape)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the footprints module.
"""

import numpy as np
import pytest
from numpy.testing import assert_equal

from photutils.utils.footprints import circular_footprint


def test_footprints():
    """
    Test circular_footprint with various radii and invalid inputs.
    """
    footprint = circular_footprint(1)
    result = np.array([[0, 1, 0],
                       [1, 1, 1],
                       [0, 1, 0]])
    assert_equal(footprint, result)

    footprint = circular_footprint(2)
    result = np.array([[0, 0, 1, 0, 0],
                       [0, 1, 1, 1, 0],
                       [1, 1, 1, 1, 1],
                       [0, 1, 1, 1, 0],
                       [0, 0, 1, 0, 0]])
    assert_equal(footprint, result)

    match = 'radius must be a positive, finite integer greater than 0'
    with pytest.raises(ValueError, match=match):
        circular_footprint(5.1)

    with pytest.raises(ValueError, match=match):
        circular_footprint(0)

    with pytest.raises(ValueError, match=match):
        circular_footprint(-1)

    with pytest.raises(ValueError, match=match):
        circular_footprint(np.inf)

    with pytest.raises(ValueError, match=match):
        circular_footprint(np.nan)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tools for detecting sources in an image.
"""

import warnings

import numpy as np
from astropy.stats import SigmaClip
from scipy.ndimage import find_objects
from scipy.ndimage import label as ndi_label

from photutils.segmentation.core import SegmentationImage
from photutils.segmentation.utils import _make_binary_structure
from photutils.utils._deprecation import deprecated_renamed_argument
from photutils.utils._parameters import (SigmaClipSentinelDefault,
                                         create_default_sigmaclip)
from photutils.utils._quantity_helpers import check_units, process_quantities
from photutils.utils._stats import nanmean, nanstd
from photutils.utils.exceptions import NoDetectionsWarning

__all__ = ['detect_sources', 'detect_threshold']


SIGMA_CLIP = SigmaClipSentinelDefault(sigma=3.0, maxiters=10)


@deprecated_renamed_argument('nsigma', 'n_sigma', '3.0', until='4.0')
def detect_threshold(data, n_sigma, *, background=None, error=None, mask=None,
                     sigma_clip=SIGMA_CLIP):
    """
    Calculate a pixel-wise threshold image that can be used to detect
    sources.

    This is a simple convenience function that uses sigma-clipped
    statistics to compute a scalar background and noise estimate. In
    general, one should perform more sophisticated estimates, e.g.,
    using `~photutils.background.Background2D`.

    Parameters
    ----------
    data : 2D `~numpy.ndarray`
        The 2D array of the image.

    n_sigma : float
        The number of standard deviations per pixel above the
        ``background`` for which to consider a pixel as possibly being
        part of a source.

    background : float or 2D `~numpy.ndarray`, optional
        The background value(s) of the input ``data``. ``background``
        may either be a scalar value or a 2D array with the same
        shape as the input ``data``. If the input ``data`` has been
        background-subtracted, then set ``background`` to ``0.0`` (this
        should be typical). If `None`, then a scalar background value
        will be estimated as the sigma-clipped image mean.

    error : float or 2D `~numpy.ndarray`, optional
        The Gaussian 1-sigma standard deviation of the background
        noise in ``data``. ``error`` should include all sources of
        "background" error, but *exclude* the Poisson error of the
        sources. If ``error`` is a 2D image, then it should represent
        the 1-sigma background error in each pixel of ``data``. If
        `None`, then a scalar background rms value will be estimated
        as the sigma-clipped image standard deviation.

    mask : 2D bool `~numpy.ndarray`, optional
        A boolean mask with the same shape as ``data``, where a `True`
        value indicates the corresponding element of ``data`` is masked.
        Masked pixels are ignored when computing the image background
        statistics.

    sigma_clip : `astropy.stats.SigmaClip` or `None`, optional
        A `~astropy.stats.SigmaClip` object that defines the sigma
        clipping parameters. If `None` then no sigma clipping will be
        performed.

    Returns
    -------
    threshold : 2D `~numpy.ndarray`
        A 2D image with the same shape (and units) as ``data``
        containing the pixel-wise threshold values.

    See Also
    --------
    :class:`photutils.background.Background2D`
    :func:`photutils.segmentation.detect_sources`
    :class:`photutils.segmentation.SourceFinder`

    Notes
    -----
    The ``mask`` and ``sigma_clip`` inputs are used only if it
    is necessary to estimate ``background`` or ``error`` using
    sigma-clipped background statistics. If ``background`` and ``error``
    are both input, then ``mask`` and ``sigma_clip`` are ignored.
    """
    inputs = (data, background, error)
    names = ('data', 'background', 'error')
    inputs, unit = process_quantities(inputs, names)
    (data, background, error) = inputs

    if sigma_clip is SIGMA_CLIP:
        sigma_clip = create_default_sigmaclip(sigma=SIGMA_CLIP.sigma,
                                              maxiters=SIGMA_CLIP.maxiters)
    if not isinstance(sigma_clip, SigmaClip):
        msg = 'sigma_clip must be a SigmaClip object'
        raise TypeError(msg)

    if background is None or error is None:
        if mask is not None:
            data = np.ma.MaskedArray(data, mask)

        clipped_data = sigma_clip(data, masked=False, return_bounds=False,
                                  copy=True)

    if background is None:
        background = nanmean(clipped_data)

    if not np.isscalar(background) and background.shape != data.shape:
        msg = ('If input background is 2D, then it must have the same '
               'shape as the input data.')
        raise ValueError(msg)

    if error is None:
        error = nanstd(clipped_data)
    if not np.isscalar(error) and error.shape != data.shape:
        msg = ('If input error is 2D, then it must have the same shape '
               'as the input data.')
        raise ValueError(msg)

    threshold = (np.broadcast_to(background, data.shape)
                 + np.broadcast_to(error * n_sigma, data.shape))

    if unit:
        threshold <<= unit

    return threshold


def _detect_sources(data, threshold, n_pixels, footprint, inverse_mask, *,
                    relabel=True, return_segmimg=True):
    """
    Detect sources above a specified threshold value in an image.

    Detected sources must have ``n_pixels`` connected pixels that are
    each greater than the ``threshold`` value in the input ``data``.

    This function is the core algorithm for detecting sources in
    an image used by `detect_sources`. This function differs from
    `detect_sources` in that it does not perform any boilerplate checks,
    it accepts a ``footprint`` argument instead of a ``connectivity``
    argument, and it accepts an ``inverse_mask`` argument instead of a
    ``mask`` argument. It is also used by the source deblending function
    for multithresholding.

    Parameters
    ----------
    data : 2D `~numpy.ndarray`
        The 2D array of the image. If filtering is desired, please input
        a convolved image.

    threshold : float or 2D `~numpy.ndarray`
        The data value or pixel-wise data values to be used for the
        detection threshold. If ``data`` is a `~astropy.units.Quantity`
        array, then ``threshold`` must have the same units as ``data``.
        A 2D ``threshold`` array must have the same shape as ``data``.

    n_pixels : int
        The minimum number of connected pixels, each greater than
        ``threshold``, that an object must have to be detected.
        ``n_pixels`` must be a positive integer.

    footprint : array_like
        A footprint that defines feature connections. As an example,
        for connectivity along pixel edges only, the footprint is
        ``np.array([[0, 1, 0]], [1, 1, 1], [0, 1, 0]])``.

    inverse_mask : 2D bool `~numpy.ndarray`
        A boolean mask, with the same shape as the input ``data``, where
        `False` values indicate masked pixels (the inverse of usual
        pixel masks). Masked pixels will not be included in any source.

    relabel : bool, optional
        If `True`, relabel the segmentation image with consecutive
        numbers.

    return_segmimg : bool, optional
        If `True`, return a `~photutils.segmentation.SegmentationImage`
        object. If `False`, return a 2D `~numpy.ndarray` segmentation
        image. The latter is used by the source deblending function.
        In that case, if only one source is found, then `None` is
        returned.

    Returns
    -------
    segment_image : `~photutils.segmentation.SegmentationImage`, \
            2D `~numpy.ndarray`, or `None`
        A 2D segmentation image, with the same shape as ``data``, where
        sources are marked by different positive integer values. A value
        of zero is reserved for the background. If ``return_segmimg``
        is `False`, then a 2D `~numpy.ndarray` segmentation image is
        returned. If no sources are found then `None` is returned.
    """
    # Ignore RuntimeWarning caused by > comparison when data contains NaNs
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', category=RuntimeWarning)
        segment_img = data > threshold

    if inverse_mask is not None:
        segment_img &= inverse_mask

    # Return None if threshold was too high to detect any sources
    if np.count_nonzero(segment_img) == 0:
        return None

    # NOTE: recasting segment_img to int and using output=segment_img
    # gives similar performance
    segment_img, nlabels = ndi_label(segment_img, structure=footprint)
    labels = np.arange(nlabels, dtype=segment_img.dtype) + 1

    # Remove objects with less than n_pixels
    # NOTE: making cutout images and setting their pixels to 0 is
    # ~10x faster than using segment_img directly and ~50% faster
    # than using ndimage.sum_labels.
    slices = find_objects(segment_img)
    segm_labels = []
    segm_slices = []
    for label, slc in zip(labels, slices, strict=True):
        cutout = segment_img[slc]
        segment_mask = (cutout == label)
        if np.count_nonzero(segment_mask) < n_pixels:
            cutout[segment_mask] = 0
            continue
        segm_labels.append(label)
        segm_slices.append(slc)

    if np.count_nonzero(segment_img) == 0:
        return None

    if relabel:
        # Relabel the segmentation image with consecutive numbers;
        # ndimage.label returns segment_img with dtype = np.int32
        # unless the input array has more than 2**31 - 1 pixels
        nlabels = len(segm_labels)
        if len(labels) != nlabels:
            label_map = np.zeros(np.max(labels) + 1,
                                 dtype=segment_img.dtype)
            labels = np.arange(nlabels, dtype=segment_img.dtype) + 1
            label_map[segm_labels] = labels
            segment_img = label_map[segment_img]
    else:
        labels = segm_labels

    if return_segmimg:
        segm = object.__new__(SegmentationImage)
        segm._data = segment_img
        segm.__dict__['labels'] = labels
        segm.__dict__['slices'] = segm_slices
        segm.__dict__['_deblend_label_map'] = {}
        return segm

    # This is used by deblend_sources
    if len(labels) == 1:
        return None

    return segment_img


@deprecated_renamed_argument('npixels', 'n_pixels', '3.0', until='4.0')
def detect_sources(data, threshold, n_pixels, *, connectivity=8, mask=None):
    """
    Detect sources above a specified threshold value in an image.

    Detected sources must have ``n_pixels`` connected pixels that are
    each greater than the ``threshold`` value in the input ``data``. The
    input ``mask`` can be used to mask pixels in the input data. Masked
    pixels will not be included in any source.

    This function does not deblend overlapping sources.
    First use this function to detect sources followed by
    :func:`~photutils.segmentation.deblend_sources` to deblend sources.
    Alternatively, use the :class:`~photutils.segmentation.SourceFinder`
    class to detect and deblend sources in a single step.

    Parameters
    ----------
    data : 2D `~numpy.ndarray`
        The 2D array of the image. If filtering is desired, please input
        a convolved image.

    threshold : float or 2D `~numpy.ndarray`
        The data value or pixel-wise data values to be used for the
        detection threshold. If ``data`` is a `~astropy.units.Quantity`
        array, then ``threshold`` must have the same units as ``data``.
        A 2D ``threshold`` array must have the same shape as ``data``.

    n_pixels : int
        The minimum number of connected pixels, each greater than
        ``threshold``, that an object must have to be detected.
        ``n_pixels`` must be a positive integer.

    connectivity : {4, 8}, optional
        The type of pixel connectivity used in determining how pixels
        are grouped into a detected source. The options are 4 or
        8 (default). 4-connected pixels touch along their edges.
        8-connected pixels touch along their edges or corners.

    mask : 2D bool `~numpy.ndarray`, optional
        A boolean mask, with the same shape as the input ``data``, where
        `True` values indicate masked pixels. Masked pixels will not be
        included in any source.

    Returns
    -------
    segment_image : `~photutils.segmentation.SegmentationImage` or `None`
        A 2D segmentation image, with the same shape as ``data``, where
        sources are marked by different positive integer values. A value
        of zero is reserved for the background. If no sources are found
        then `None` is returned.

    Raises
    ------
    NoDetectionsWarning
        If no sources are found.

    See Also
    --------
    :func:`photutils.segmentation.deblend_sources`
    :class:`photutils.segmentation.SourceFinder`

    Examples
    --------
    .. plot::
        :include-source:

        import matplotlib.pyplot as plt
        from astropy.convolution import convolve
        from astropy.visualization import simple_norm
        from photutils.background import Background2D, MedianBackground
        from photutils.datasets import make_100gaussians_image
        from photutils.segmentation import (detect_sources,
                                            make_2dgaussian_kernel)

        # Make a simulated image
        data = make_100gaussians_image()

        # Estimate the background using Background2D and subtract it
        bkg_estimator = MedianBackground()
        bkg = Background2D(data, (50, 50), filter_size=(3, 3),
                           bkg_estimator=bkg_estimator)
        data -= bkg.background  # subtract the background

        # Convolve the data
        kernel = make_2dgaussian_kernel(3.0, size=5)
        convolved_data = convolve(data, kernel)

        # Detect the sources
        threshold = 1.5 * bkg.background_rms  # set the detection threshold
        segment_map = detect_sources(convolved_data, threshold, n_pixels=10)

        # Plot the image and the segmentation image
        fig, (ax1, ax2) = plt.subplots(2, 1, figsize=(8, 10))
        norm = simple_norm(data, 'sqrt', percent=99.5)
        ax1.imshow(data, norm=norm, origin='lower')
        segment_map.imshow(ax=ax2)
        fig.tight_layout()
    """
    check_units((data, threshold), ('data', 'threshold'))

    if (n_pixels <= 0) or (int(n_pixels) != n_pixels):
        msg = f'n_pixels must be a positive integer, got {n_pixels!r}'
        raise ValueError(msg)

    if mask is not None:
        if mask.shape != data.shape:
            msg = 'mask must have the same shape as the input image'
            raise ValueError(msg)
        if mask.all():
            msg = ('mask must not be True for every pixel. There are no '
                   'unmasked pixels in the image to detect sources.')
            raise ValueError(msg)
        inverse_mask = np.logical_not(mask)
    else:
        inverse_mask = None

    footprint = _make_binary_structure(data.ndim, connectivity)

    segm = _detect_sources(data, threshold, n_pixels, footprint,
                           inverse_mask, relabel=True, return_segmimg=True)

    if segm is None:
        msg = ('No sources were found. Try lowering the threshold or '
               'pixels parameters.')
        warnings.warn(msg, NoDetectionsWarning)

    return segm



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Class for aperture masks.
"""

import warnings

import astropy.units as u
import numpy as np

from photutils.utils._deprecation import deprecated_positional_kwargs

__all__ = ['ApertureMask']


class ApertureMask:
    """
    Class for an aperture mask.

    Parameters
    ----------
    data : array_like
        A 2D array representing the fractional overlap of an aperture
        on the pixel grid. This should be the full-sized (i.e., not
        truncated) array that is the direct output of one of the
        low-level `photutils.geometry` functions.

    bbox : `photutils.aperture.BoundingBox`
        The bounding box object defining the aperture minimal bounding
        box.
    """

    def __init__(self, data, bbox):
        self.data = np.asanyarray(data)
        if self.data.shape != bbox.shape:
            msg = 'mask data and bounding box must have the same shape'
            raise ValueError(msg)
        self.bbox = bbox
        self._mask = (self.data == 0)

    # NumPy calls `obj.__array__(dtype)` positionally with
    # `np.asarray(obj, dtype=int)`, so dtype must remain a positional
    # argument.
    def __array__(self, dtype=None, *, copy=None):
        """
        Array representation of the mask data array (e.g., for
        matplotlib).
        """
        return np.array(self.data, dtype=dtype, copy=copy)

    @property
    def shape(self):
        """
        The shape of the mask data array.
        """
        return self.data.shape

    def get_overlap_slices(self, shape):
        """
        Get slices for the overlapping part of the aperture mask and a
        2D array.

        Parameters
        ----------
        shape : 2-tuple of int
            The shape of the 2D array.

        Returns
        -------
        slices_large : tuple of slices or `None`
            A tuple of slice objects for each axis of the large array,
            such that ``large_array[slices_large]`` extracts the region
            of the large array that overlaps with the small array.
            `None` is returned if there is no overlap of the bounding
            box with the given image shape.

        slices_small : tuple of slices or `None`
            A tuple of slice objects for each axis of the aperture mask
            array such that ``small_array[slices_small]`` extracts the
            region that is inside the large array. `None` is returned if
            there is no overlap of the bounding box with the given image
            shape.
        """
        return self.bbox.get_overlap_slices(shape)

    @deprecated_positional_kwargs(since='3.0', until='4.0')
    def to_image(self, shape, dtype=float):
        """
        Return an image of the mask in a 2D array of the given shape,
        taking any edge effects into account.

        Parameters
        ----------
        shape : tuple of int
            The ``(ny, nx)`` shape of the output array.

        dtype : data-type, optional
            The desired data type for the array. This should be a
            floating data type if the `ApertureMask` was created with
            the "exact" or "subpixel" method, otherwise the fractional
            mask weights will be altered. An integer data type may be
            used if the `ApertureMask` was created with the "center"
            method.

        Returns
        -------
        result : `~numpy.ndarray`
            A 2D array of the mask.
        """
        if len(shape) != 2:
            msg = 'input shape must have 2 elements'
            raise ValueError(msg)

        # Find the overlap of the mask on the output image shape
        slices_large, slices_small = self.get_overlap_slices(shape)

        if slices_small is None:
            return None  # no overlap

        # Insert the mask into the output image
        image = np.zeros(shape, dtype=dtype)
        image[slices_large] = self.data[slices_small]
        return image

    @deprecated_positional_kwargs(since='3.0', until='4.0')
    def cutout(self, data, fill_value=0.0, copy=False):
        """
        Create a cutout from the input data over the mask bounding box,
        taking any edge effects into account.

        Parameters
        ----------
        data : array_like
            A 2D array on which to apply the aperture mask.

        fill_value : float, optional
            The value used to fill pixels where the aperture mask does
            not overlap with the input ``data``. The default is 0.

        copy : bool, optional
            If `True` then the returned cutout array will always be hold
            a copy of the input ``data``. If `False` and the mask is
            fully within the input ``data``, then the returned cutout
            array will be a view into the input ``data``. In cases where
            the mask partially overlaps or has no overlap with the input
            ``data``, the returned cutout array will always hold a copy
            of the input ``data`` (i.e., this keyword has no effect).

        Returns
        -------
        result : `~numpy.ndarray` or `None`
            A 2D array cut out from the input ``data`` representing
            the same cutout region as the aperture mask. If there is a
            partial overlap of the aperture mask with the input data,
            pixels outside the data will be assigned to ``fill_value``.
            `None` is returned if there is no overlap of the aperture
            with the input ``data``.
        """
        data = np.asanyarray(data)
        if data.ndim != 2:
            msg = 'data must be a 2D array'
            raise ValueError(msg)

        # Find the overlap of the mask on the output image shape
        slices_large, slices_small = self.get_overlap_slices(data.shape)

        if slices_small is None:
            return None  # no overlap

        cutout_shape = (slices_small[0].stop - slices_small[0].start,
                        slices_small[1].stop - slices_small[1].start)

        if cutout_shape == self.shape:
            cutout = data[slices_large]
            if copy:
                cutout = np.copy(cutout)
            return cutout

        # Cutout is always a copy for partial overlap
        dtype = float if ~np.isfinite(fill_value) else data.dtype
        cutout = np.zeros(self.shape, dtype=dtype)
        cutout[:] = fill_value
        cutout[slices_small] = data[slices_large]

        if isinstance(data, u.Quantity):
            cutout <<= data.unit

        return cutout

    @deprecated_positional_kwargs(since='3.0', until='4.0')
    def multiply(self, data, fill_value=0.0):
        """
        Multiply the aperture mask with the input data, taking any edge
        effects into account.

        The result is a mask-weighted cutout from the data.

        Parameters
        ----------
        data : array_like or `~astropy.units.Quantity`
            The 2D array to multiply with the aperture mask.

        fill_value : float, optional
            The value is used to fill pixels where the aperture mask
            does not overlap with the input ``data``. The default is 0.

        Returns
        -------
        result : `~numpy.ndarray` or `None`
            A 2D mask-weighted cutout from the input ``data``. If
            there is a partial overlap of the aperture mask with the
            input data, pixels outside the data will be assigned to
            ``fill_value`` before being multiplied with the mask. `None`
            is returned if there is no overlap of the aperture with the
            input ``data``.
        """
        cutout = self.cutout(data, fill_value=fill_value)
        if cutout is None:
            return None

        # Ignore multiplication with non-finite data values
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', RuntimeWarning)
            weighted_cutout = cutout * self.data

        # Fill values outside the mask but within the bounding box
        weighted_cutout[self._mask] = fill_value

        return weighted_cutout

    def _get_overlap_cutouts(self, shape, *, mask=None):
        """
        Get the aperture mask weights, pixel mask, and slice for the
        overlap with the input shape.

        If input, the ``mask`` is included in the output pixel mask
        cutout.

        Parameters
        ----------
        shape : tuple of int
            The shape of data.

        mask : array_like (bool), optional
            A boolean mask with the same shape as ``shape`` where a
            `True` value indicates a masked pixel.

        Returns
        -------
        slices_large : tuple of slices or `None`
            A tuple of slice objects for each axis of the large array
            of given ``shape``, such that ``large_array[slices_large]``
            extracts the region of the large array that overlaps with
            the small array. `None` is returned if there is no overlap
            of the bounding box with the given image shape.

        aper_weights: 2D float `~numpy.ndarray`
            The cutout aperture mask weights for the overlap.

        pixel_mask: 2D bool `~numpy.ndarray`
            The cutout pixel mask for the overlap.

        Notes
        -----
        This method is separate from ``get_values`` to facilitate
        applying the same slices, aper_weights, and pixel_mask to
        multiple associated arrays (e.g., data and error arrays). It is
        used in this way by the `PixelAperture.do_photometry` method.
        """
        if mask is not None and mask.shape != shape:
            msg = 'mask and data must have the same shape'
            raise ValueError(msg)

        slc_large, slc_small = self.get_overlap_slices(shape)
        if slc_large is None:  # no overlap
            return None, None, None

        aper_weights = self.data[slc_small]
        pixel_mask = (aper_weights > 0)  # good pixels

        if mask is not None:
            pixel_mask &= ~mask[slc_large]

        return slc_large, aper_weights, pixel_mask

    @deprecated_positional_kwargs(since='3.0', until='4.0')
    def get_values(self, data, mask=None):
        """
        Get the mask-weighted pixel values from the data as a 1D array.

        If the ``ApertureMask`` was created with ``method='center'``,
        (where the mask weights are only 1 or 0), then the returned
        values will simply be pixel values extracted from the data.

        Parameters
        ----------
        data : array_like or `~astropy.units.Quantity`
            The 2D array from which to get mask-weighted values.

        mask : array_like (bool), optional
            A boolean mask with the same shape as ``data`` where a
            `True` value indicates the corresponding element of ``data``
            is not returned in the result.

        Returns
        -------
        result : `~numpy.ndarray`
            A 1D array of mask-weighted pixel values from the input
            ``data``. If there is no overlap of the aperture with the
            input ``data``, the result will be an empty array with shape
            (0,).
        """
        slc_large, aper_weights, pixel_mask = self._get_overlap_cutouts(
            data.shape, mask=mask)

        if slc_large is None:
            return np.array([])

        # Ignore multiplication with non-finite data values
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', RuntimeWarning)
            # pixel_mask is used so that pixels value where data = 0 and
            # aper_weights != 0 are still returned
            return (data[slc_large] * aper_weights)[pixel_mask]



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the _wcs_helpers module.
"""

import astropy.units as u
import numpy as np
import pytest
from astropy.coordinates import Angle, SkyCoord
from numpy.testing import assert_allclose

from photutils.utils._wcs_helpers import (compute_local_wcs_jacobian,
                                          jacobian_pixel_to_sky_mean_scale,
                                          jacobian_pixel_to_sky_scales,
                                          jacobian_sky_to_pixel_mean_scale,
                                          jacobian_sky_to_pixel_scales,
                                          pixel_ellipse_to_sky_svd,
                                          pixel_to_sky_mean_scale,
                                          pixel_to_sky_scales,
                                          pixel_to_sky_svd_scales,
                                          sky_ellipse_to_pixel_svd,
                                          sky_to_pixel_mean_scale,
                                          sky_to_pixel_scales,
                                          sky_to_pixel_svd_scales,
                                          wcs_pixel_scale_angle)
from photutils.utils.tests.conftest import WCS_CDELT_ARCSEC, WCS_CENTER


@pytest.fixture
def center_xy_coord(simple_wcs):
    """
    Return the center (x, y) tuple at CRPIX of the simple WCS.
    """
    x, y = simple_wcs.world_to_pixel(WCS_CENTER)
    return (x, y)


class TestComputeLocalWCSJacobian:
    """
    Tests for `compute_local_wcs_jacobian`.
    """

    def test_shape(self, simple_wcs):
        """
        The Jacobian must be a 2x2 array.
        """
        jac = compute_local_wcs_jacobian(WCS_CENTER, simple_wcs)
        assert jac.shape == (2, 2)

    def test_simple_wcs_diagonal(self, simple_wcs):
        """
        For an axis-aligned TAN WCS the Jacobian should be nearly
        diagonal with magnitudes ~ 1/WCS_CDELT_ARCSEC.
        """
        jac = compute_local_wcs_jacobian(WCS_CENTER, simple_wcs)
        # Off-diagonal elements should be near zero
        assert_allclose(jac[0, 1], 0.0, atol=1e-4)
        assert_allclose(jac[1, 0], 0.0, atol=1e-4)
        # Diagonal: RA axis (pix/arcsec) is negative (RA increases left)
        expected_scale = 1.0 / WCS_CDELT_ARCSEC
        assert_allclose(np.abs(jac[0, 0]), expected_scale)
        assert_allclose(np.abs(jac[1, 1]), expected_scale)

    def test_rotated_wcs(self, rotated_wcs):
        """
        For a rotated WCS the off-diagonal elements should be nonzero,
        but the singular values should still match 1/WCS_CDELT_ARCSEC.
        """
        jac = compute_local_wcs_jacobian(WCS_CENTER, rotated_wcs)
        sv = np.linalg.svd(jac, compute_uv=False)
        expected_scale = 1.0 / WCS_CDELT_ARCSEC
        assert_allclose(sv, expected_scale, rtol=1e-6)

    def test_sip_wcs(self, sip_wcs):
        """
        For a SIP WCS the Jacobian should still be close to the
        undistorted value near the reference pixel.
        """
        jac = compute_local_wcs_jacobian(WCS_CENTER, sip_wcs)
        sv = np.linalg.svd(jac, compute_uv=False)
        expected_scale = 1.0 / WCS_CDELT_ARCSEC
        assert_allclose(sv, expected_scale, rtol=1e-5)

    def test_inverse_of_forward(self, simple_wcs):
        """
        The Jacobian should be the inverse of the forward
        d(sky)/d(pixel) matrix derived from 1-pixel offsets.
        """
        jac = compute_local_wcs_jacobian(WCS_CENTER, simple_wcs)
        # A 1-pixel step should map to ~CDELT arcsec in sky
        forward = np.linalg.inv(jac)
        # Diagonal magnitudes should be ~WCS_CDELT_ARCSEC
        assert_allclose(np.abs(forward[0, 0]), WCS_CDELT_ARCSEC)
        assert_allclose(np.abs(forward[1, 1]), WCS_CDELT_ARCSEC)

    def test_determinant_sign(self, simple_wcs):
        """
        Standard WCS (RA increasing to the left) should have negative
        determinant.
        """
        jac = compute_local_wcs_jacobian(WCS_CENTER, simple_wcs)
        assert np.linalg.det(jac) < 0


class TestWcsPixelScaleAngle:
    """
    Tests for `wcs_pixel_scale_angle`.
    """

    def test_return_types(self, simple_wcs):
        """
        Should return (tuple, float, Angle).
        """
        xy_coord, scale, angle = wcs_pixel_scale_angle(
            WCS_CENTER, simple_wcs)
        assert isinstance(xy_coord, tuple)
        assert isinstance(scale, float)
        assert isinstance(angle, Angle)

    def test_simple_wcs_scale(self, simple_wcs):
        """
        For a simple TAN WCS, scale should equal CDELT in arcsec/pixel.
        """
        _, scale, _ = wcs_pixel_scale_angle(WCS_CENTER, simple_wcs)
        assert_allclose(scale, WCS_CDELT_ARCSEC)

    def test_simple_wcs_angle(self, simple_wcs):
        """
        For an axis-aligned TAN WCS with CDELT=[-c, c], North is along
        +y, so the angle should be ~90 degrees.
        """
        _, _, angle = wcs_pixel_scale_angle(WCS_CENTER, simple_wcs)
        assert_allclose(angle.deg, 90.0)

    def test_angle_wrapped(self, simple_wcs):
        """
        The angle should be in [0, 360) degrees.
        """
        _, _, angle = wcs_pixel_scale_angle(WCS_CENTER, simple_wcs)
        assert 0.0 <= angle.deg < 360.0

    def test_rotated_wcs_angle(self, rotated_wcs):
        """
        For a 25-degree rotated WCS, the North angle should shift by
        ~25 degrees from the axis-aligned value (~90 deg).
        """
        _, _, angle = wcs_pixel_scale_angle(WCS_CENTER, rotated_wcs)
        # The rotation should be about 90 - 25 = 65 degrees
        assert_allclose(angle.deg, 90.0 - 25.0)

    def test_rotated_wcs_scale(self, rotated_wcs):
        """
        Rotation should not change the pixel scale.
        """
        _, scale, _ = wcs_pixel_scale_angle(WCS_CENTER, rotated_wcs)
        assert_allclose(scale, WCS_CDELT_ARCSEC)

    def test_nonsquare_wcs_scale(self, nonsquare_wcs):
        """
        For non-square pixels the scale should be the geometric mean.
        """
        _, scale, _ = wcs_pixel_scale_angle(WCS_CENTER, nonsquare_wcs)
        expected = np.sqrt(0.03 * 0.05) * 3600
        assert_allclose(scale, expected, rtol=1e-5)

    def test_pixel_coordinate(self, simple_wcs):
        """
        The returned xy_coord should match world_to_pixel.
        """
        xy_coord, _, _ = wcs_pixel_scale_angle(WCS_CENTER, simple_wcs)
        x_exp, y_exp = simple_wcs.world_to_pixel(WCS_CENTER)
        assert_allclose(xy_coord[0], x_exp)
        assert_allclose(xy_coord[1], y_exp)

    def test_off_center_position(self, simple_wcs):
        """
        Test a position away from the WCS reference pixel.
        """
        skycoord = SkyCoord(100.5 * u.deg, 30.5 * u.deg)
        _, scale, angle = wcs_pixel_scale_angle(skycoord, simple_wcs)
        assert scale > 0
        assert 0.0 <= angle.deg < 360.0


class TestJacobianDirectionalScales:
    """
    Tests for `jacobian_sky_to_pixel_scales` and
    `jacobian_pixel_to_sky_scales`.
    """

    @pytest.mark.parametrize('sky_angle_deg', [0, 30, 90, 150, 270])
    def test_sky_to_pixel_return_types(self, simple_wcs,
                                       sky_angle_deg):
        """
        Should return (tuple, float, float, Angle).
        """
        sky_angle_rad = np.radians(sky_angle_deg)
        pix_position, sw, sh, pangle = jacobian_sky_to_pixel_scales(
            WCS_CENTER, simple_wcs, sky_angle_rad)
        assert isinstance(pix_position, tuple)
        assert isinstance(sw, (float, np.floating))
        assert isinstance(sh, (float, np.floating))
        assert isinstance(pangle, Angle)

    @pytest.mark.parametrize('pixel_angle_deg', [0, 45, 90, 180, 315])
    def test_pixel_to_sky_return_types(self, simple_wcs, center_xy_coord,
                                       pixel_angle_deg):
        """
        Should return (SkyCoord, float, float, Angle).
        """
        pixel_angle_rad = np.radians(pixel_angle_deg)
        sky_position, sw, sh, sangle = jacobian_pixel_to_sky_scales(
            center_xy_coord, simple_wcs, pixel_angle_rad)
        assert isinstance(sky_position, SkyCoord)
        assert isinstance(sw, (float, np.floating))
        assert isinstance(sh, (float, np.floating))
        assert isinstance(sangle, Angle)

    def test_sky_to_pixel_simple_scales(self, simple_wcs):
        """
        For a simple WCS, the directional scale factors should both
        equal 1/WCS_CDELT_ARCSEC (pixels per arcsec).
        """
        _, sw, sh, _ = jacobian_sky_to_pixel_scales(
            WCS_CENTER, simple_wcs, 0.0)
        expected = 1.0 / WCS_CDELT_ARCSEC
        assert_allclose(sw, expected)
        assert_allclose(sh, expected)

    def test_pixel_to_sky_simple_scales(self, simple_wcs, center_xy_coord):
        """
        For a simple WCS, the directional scale factors should both
        equal WCS_CDELT_ARCSEC (arcsec per pixel).
        """
        _, sw, sh, _ = jacobian_pixel_to_sky_scales(
            center_xy_coord, simple_wcs, 0.0)
        assert_allclose(sw, WCS_CDELT_ARCSEC)
        assert_allclose(sh, WCS_CDELT_ARCSEC)

    def test_pixel_angle_wrapped(self, simple_wcs):
        """
        The pixel angle must be in [0, 360) degrees.
        """
        _, _, _, pangle = jacobian_sky_to_pixel_scales(
            WCS_CENTER, simple_wcs, 0.0)
        assert 0.0 <= pangle.deg < 360.0

    def test_sky_angle_wrapped(self, simple_wcs, center_xy_coord):
        """
        The sky angle must be in [0, 360) degrees.
        """
        _, _, _, sangle = jacobian_pixel_to_sky_scales(
            center_xy_coord, simple_wcs, 0.0)
        assert 0.0 <= sangle.deg < 360.0

    @pytest.mark.parametrize('angle_deg', [0, 30, 90, 270])
    def test_roundtrip_angle(self, simple_wcs, angle_deg):
        """
        Converting sky -> pixel -> sky should recover the original
        angle.
        """
        sky_angle_rad = np.radians(angle_deg)
        center_pix, _, _, pangle = jacobian_sky_to_pixel_scales(
            WCS_CENTER, simple_wcs, sky_angle_rad)
        _, _, _, sangle = jacobian_pixel_to_sky_scales(
            center_pix, simple_wcs, pangle.rad)
        assert_allclose(sangle.deg % 360, angle_deg % 360, atol=1e-6)

    @pytest.mark.parametrize('angle_deg', [0, 45, 135])
    def test_roundtrip_scales(self, simple_wcs, angle_deg):
        """
        Converting sky -> pixel -> sky should recover the original
        scale factors.
        """
        sky_angle_rad = np.radians(angle_deg)
        center_pix, sw, sh, pangle = jacobian_sky_to_pixel_scales(
            WCS_CENTER, simple_wcs, sky_angle_rad)
        _, sw_rt, sh_rt, _ = jacobian_pixel_to_sky_scales(
            center_pix, simple_wcs, pangle.rad)
        # sw (pix/arcsec) * sw_rt (arcsec/pix) ~ 1
        assert_allclose(sw * sw_rt, 1.0)
        assert_allclose(sh * sh_rt, 1.0)

    def test_sip_wcs_positive_scales(self, sip_wcs):
        """
        Scale factors should be positive even for distorted WCS.
        """
        _, sw, sh, _ = jacobian_sky_to_pixel_scales(
            WCS_CENTER, sip_wcs, np.radians(30))
        assert sw > 0
        assert sh > 0

    def test_rotated_wcs(self, rotated_wcs):
        """
        For a rotated (but isotropic) WCS, scales should still equal
        1/WCS_CDELT_ARCSEC.
        """
        _, sw, sh, _ = jacobian_sky_to_pixel_scales(
            WCS_CENTER, rotated_wcs, 0.0)
        expected = 1.0 / WCS_CDELT_ARCSEC
        assert_allclose(sw, expected, rtol=1e-6)
        assert_allclose(sh, expected, rtol=1e-6)


class TestJacobianMeanScale:
    """
    Tests for `jacobian_sky_to_pixel_mean_scale` and
    `jacobian_pixel_to_sky_mean_scale`.
    """

    def test_sky_to_pixel_return_types(self, simple_wcs):
        """
        Should return (tuple, float).
        """
        pix_position, scale = jacobian_sky_to_pixel_mean_scale(
            WCS_CENTER, simple_wcs)
        assert isinstance(pix_position, tuple)
        assert isinstance(scale, (float, np.floating))

    def test_pixel_to_sky_return_types(self, simple_wcs, center_xy_coord):
        """
        Should return (SkyCoord, float).
        """
        sky_position, scale = jacobian_pixel_to_sky_mean_scale(
            center_xy_coord, simple_wcs)
        assert isinstance(sky_position, SkyCoord)
        assert isinstance(scale, (float, np.floating))

    def test_sky_to_pixel_simple_scale(self, simple_wcs):
        """
        For an isotropic WCS, the mean scale should equal
        1/WCS_CDELT_ARCSEC.
        """
        _, scale = jacobian_sky_to_pixel_mean_scale(
            WCS_CENTER, simple_wcs)
        assert_allclose(scale, 1.0 / WCS_CDELT_ARCSEC)

    def test_pixel_to_sky_simple_scale(self, simple_wcs, center_xy_coord):
        """
        For an isotropic WCS, the mean scale should equal WCS_CDELT_ARCSEC.
        """
        _, scale = jacobian_pixel_to_sky_mean_scale(
            center_xy_coord, simple_wcs)
        assert_allclose(scale, WCS_CDELT_ARCSEC)

    def test_roundtrip_scale(self, simple_wcs):
        """
        Sky -> pixel mean_scale * pixel -> sky mean_scale should ~ 1.
        """
        center_pix, s2p = jacobian_sky_to_pixel_mean_scale(
            WCS_CENTER, simple_wcs)
        _, p2s = jacobian_pixel_to_sky_mean_scale(center_pix, simple_wcs)
        assert_allclose(s2p * p2s, 1.0)

    def test_sip_wcs_positive(self, sip_wcs):
        """
        Mean scale should be positive for distorted WCS.
        """
        _, scale = jacobian_sky_to_pixel_mean_scale(
            WCS_CENTER, sip_wcs)
        assert scale > 0

    def test_center_coordinates(self, simple_wcs):
        """
        The returned pix_position should match world_to_pixel.
        """
        pix_position, _ = jacobian_sky_to_pixel_mean_scale(
            WCS_CENTER, simple_wcs)
        x_exp, y_exp = simple_wcs.world_to_pixel(WCS_CENTER)
        assert_allclose(pix_position[0], x_exp)
        assert_allclose(pix_position[1], y_exp)


class TestDispatchScales:
    """
    Tests for `sky_to_pixel_scales` and `pixel_to_sky_scales` dispatch
    helpers that route between offset and Jacobian methods based on
    ``wcs.has_distortion``.
    """

    def test_no_distortion_dispatches_offset(self, simple_wcs):
        """
        For a non-distorted WCS, the dispatch helper should produce
        equal w/h scales (isotropic offset method).
        """
        assert not simple_wcs.has_distortion
        _, sw, sh, _ = sky_to_pixel_scales(WCS_CENTER, simple_wcs, 0.0)
        assert_allclose(sw, sh)

    def test_distortion_dispatches_jacobian(self, sip_wcs):
        """
        For a SIP WCS, the dispatch helper should use the Jacobian path.
        """
        assert sip_wcs.has_distortion
        _, sw, sh, _ = sky_to_pixel_scales(
            WCS_CENTER, sip_wcs, np.radians(30))
        # Scales should be close but not necessarily identical
        assert sw > 0
        assert sh > 0

    @pytest.mark.parametrize('angle_deg', [0, 45, 90, 180])
    def test_sky_to_pixel_return_types(self, simple_wcs,
                                       angle_deg):
        """
        Should return (tuple, float, float, Angle).
        """
        pix_position, _sw, _sh, pangle = sky_to_pixel_scales(
            WCS_CENTER, simple_wcs, np.radians(angle_deg))
        assert isinstance(pix_position, tuple)
        assert isinstance(pangle, Angle)
        assert 0.0 <= pangle.deg < 360.0

    @pytest.mark.parametrize('angle_deg', [0, 45, 90, 180])
    def test_pixel_to_sky_return_types(self, simple_wcs, center_xy_coord,
                                       angle_deg):
        """
        Should return (SkyCoord, float, float, Angle).
        """
        sky_position, _sw, _sh, sangle = pixel_to_sky_scales(
            center_xy_coord, simple_wcs, np.radians(angle_deg))
        assert isinstance(sky_position, SkyCoord)
        assert isinstance(sangle, Angle)
        assert 0.0 <= sangle.deg < 360.0

    @pytest.mark.parametrize('angle_deg', [0, 30, 90, 270])
    def test_roundtrip_simple_wcs(self, simple_wcs,
                                  angle_deg):
        """
        Converting sky -> pixel -> sky with a simple WCS should recover
        the original angle.
        """
        sky_angle_rad = np.radians(angle_deg)
        center_pix, _, _, pangle = sky_to_pixel_scales(
            WCS_CENTER, simple_wcs, sky_angle_rad)
        _, _, _, sangle = pixel_to_sky_scales(
            center_pix, simple_wcs, pangle.rad)
        assert_allclose(sangle.deg % 360, angle_deg % 360, atol=1e-6)

    @pytest.mark.parametrize('angle_deg', [0, 45, 90])
    def test_roundtrip_sip_wcs(self, sip_wcs, angle_deg):
        """
        Roundtrip through the Jacobian path should recover the original
        angle within tolerance.
        """
        sky_angle_rad = np.radians(angle_deg)
        center_pix, _, _, pangle = sky_to_pixel_scales(
            WCS_CENTER, sip_wcs, sky_angle_rad)
        _, _, _, sangle = pixel_to_sky_scales(
            center_pix, sip_wcs, pangle.rad)
        assert_allclose(sangle.deg % 360, angle_deg % 360, atol=1e-6)

    @pytest.mark.parametrize('angle_deg', [0, 45, 90])
    def test_roundtrip_rotated_wcs(self, rotated_wcs,
                                   angle_deg):
        """
        Roundtrip through a rotated WCS should recover the original
        angle.
        """
        sky_angle_rad = np.radians(angle_deg)
        center_pix, _, _, pangle = sky_to_pixel_scales(
            WCS_CENTER, rotated_wcs, sky_angle_rad)
        _, _, _, sangle = pixel_to_sky_scales(
            center_pix, rotated_wcs, pangle.rad)
        diff = abs(sangle.deg - angle_deg) % 360
        assert min(diff, 360 - diff) < 0.5

    def test_simple_wcs_scale_value(self, simple_wcs):
        """
        For a simple WCS, scales should be 1/WCS_CDELT_ARCSEC.
        """
        _, sw, sh, _ = sky_to_pixel_scales(
            WCS_CENTER, simple_wcs, 0.0)
        expected = 1.0 / WCS_CDELT_ARCSEC
        assert_allclose(sw, expected)
        assert_allclose(sh, expected)

    def test_pixel_to_sky_scale_value(self, simple_wcs, center_xy_coord):
        """
        For a simple WCS, scales should be WCS_CDELT_ARCSEC.
        """
        _, sw, sh, _ = pixel_to_sky_scales(
            center_xy_coord, simple_wcs, 0.0)
        assert_allclose(sw, WCS_CDELT_ARCSEC)
        assert_allclose(sh, WCS_CDELT_ARCSEC)

    def test_zero_angle(self, simple_wcs):
        """
        A zero sky angle should work without error.
        """
        _center, sw, sh, _pangle = sky_to_pixel_scales(
            WCS_CENTER, simple_wcs, 0.0)
        assert sw > 0
        assert sh > 0

    def test_two_pi_angle(self, simple_wcs):
        """
        An angle of 2*pi (360 degrees) should be equivalent to 0.
        """
        _, sw_0, sh_0, pa_0 = sky_to_pixel_scales(
            WCS_CENTER, simple_wcs, 0.0)
        _, sw_2pi, sh_2pi, pa_2pi = sky_to_pixel_scales(
            WCS_CENTER, simple_wcs, 2 * np.pi)
        assert_allclose(sw_0, sw_2pi, rtol=1e-10)
        assert_allclose(sh_0, sh_2pi, rtol=1e-10)
        assert_allclose(pa_0.deg % 360, pa_2pi.deg % 360, atol=1e-6)

    def test_negative_angle(self, simple_wcs):
        """
        A negative sky angle should be handled correctly.
        """
        _, sw, sh, _pangle = sky_to_pixel_scales(
            WCS_CENTER, simple_wcs, -np.pi / 4)
        assert sw > 0
        assert sh > 0

    def test_consistency_offset_jacobian(self, simple_wcs):
        """
        For a simple WCS (no distortion), the offset method and the
        Jacobian method should give consistent results.
        """
        sky_angle_rad = np.radians(30.0)

        # Offset path (via dispatch)
        c1, sw1, sh1, pa1 = sky_to_pixel_scales(
            WCS_CENTER, simple_wcs, sky_angle_rad)

        # Jacobian path (direct call)
        c2, sw2, sh2, pa2 = jacobian_sky_to_pixel_scales(
            WCS_CENTER, simple_wcs, sky_angle_rad)

        assert_allclose(c1[0], c2[0])
        assert_allclose(c1[1], c2[1])
        assert_allclose(sw1, sw2)
        assert_allclose(sh1, sh2)
        assert_allclose(pa1.deg % 360, pa2.deg % 360, atol=1e-5)

    def test_consistency_pixel_offset_jacobian(self, simple_wcs,
                                               center_xy_coord):
        """
        For a simple WCS, both paths should give consistent pixel ->
        sky results.
        """
        pixel_angle_rad = np.radians(45.0)

        # Offset path (via dispatch)
        c1, sw1, sh1, sa1 = pixel_to_sky_scales(
            center_xy_coord, simple_wcs, pixel_angle_rad)

        # Jacobian path (direct call)
        c2, sw2, sh2, sa2 = jacobian_pixel_to_sky_scales(
            center_xy_coord, simple_wcs, pixel_angle_rad)

        assert_allclose(c1.ra.deg, c2.ra.deg)
        assert_allclose(c1.dec.deg, c2.dec.deg)
        assert_allclose(sw1, sw2)
        assert_allclose(sh1, sh2)
        assert_allclose(sa1.deg % 360, sa2.deg % 360, atol=1e-5)


class TestDispatchMeanScale:
    """
    Tests for `sky_to_pixel_mean_scale` and `pixel_to_sky_mean_scale`
    dispatch helpers.
    """

    def test_no_distortion_returns(self, simple_wcs):
        """
        Should return (tuple, float) for non-distorted WCS.
        """
        pix_position, scale = sky_to_pixel_mean_scale(WCS_CENTER, simple_wcs)
        assert isinstance(pix_position, tuple)
        assert isinstance(scale, float)

    def test_distortion_returns(self, sip_wcs):
        """
        Should return (tuple, float/np.floating) for distorted WCS.
        """
        pix_position, scale = sky_to_pixel_mean_scale(WCS_CENTER, sip_wcs)
        assert isinstance(pix_position, tuple)
        assert isinstance(scale, (float, np.floating))

    def test_no_distortion_scale(self, simple_wcs):
        """
        For a simple WCS, mean scale should be 1/WCS_CDELT_ARCSEC.
        """
        _, scale = sky_to_pixel_mean_scale(WCS_CENTER, simple_wcs)
        assert_allclose(scale, 1.0 / WCS_CDELT_ARCSEC)

    def test_distortion_scale(self, sip_wcs):
        """
        For a SIP WCS near the reference pixel, the mean scale should be
        close to the undistorted value.
        """
        _, scale = sky_to_pixel_mean_scale(WCS_CENTER, sip_wcs)
        assert_allclose(scale, 1.0 / WCS_CDELT_ARCSEC, rtol=1e-6)

    def test_pixel_to_sky_no_distortion(self, simple_wcs, center_xy_coord):
        """
        For a simple WCS, pixel_to_sky mean scale should be
        WCS_CDELT_ARCSEC.
        """
        sky_position, scale = pixel_to_sky_mean_scale(
            center_xy_coord, simple_wcs)
        assert isinstance(sky_position, SkyCoord)
        assert_allclose(scale, WCS_CDELT_ARCSEC)

    def test_pixel_to_sky_distortion(self, sip_wcs):
        """
        For a SIP WCS, pixel_to_sky mean scale should be close to
        WCS_CDELT_ARCSEC near the reference pixel.
        """
        xy_coord = (9.5, 9.5)
        sky_position, scale = pixel_to_sky_mean_scale(xy_coord, sip_wcs)
        assert isinstance(sky_position, SkyCoord)
        assert_allclose(scale, WCS_CDELT_ARCSEC, rtol=1e-6)

    def test_roundtrip(self, simple_wcs):
        """
        Sky -> pixel mean_scale * pixel -> sky mean_scale should ~ 1.
        """
        center_pix, s2p = sky_to_pixel_mean_scale(
            WCS_CENTER, simple_wcs)
        _, p2s = pixel_to_sky_mean_scale(center_pix, simple_wcs)
        assert_allclose(s2p * p2s, 1.0)

    def test_roundtrip_sip(self, sip_wcs):
        """
        Roundtrip with SIP WCS should give product ~ 1.
        """
        center_pix, s2p = sky_to_pixel_mean_scale(
            WCS_CENTER, sip_wcs)
        _, p2s = pixel_to_sky_mean_scale(center_pix, sip_wcs)
        assert_allclose(s2p * p2s, 1.0)

    def test_consistency_offset_jacobian(self, simple_wcs):
        """
        For a simple WCS, both mean-scale paths should agree.
        """
        # Offset path (via dispatch)
        c1, s1 = sky_to_pixel_mean_scale(WCS_CENTER, simple_wcs)

        # Jacobian path (direct call)
        c2, s2 = jacobian_sky_to_pixel_mean_scale(
            WCS_CENTER, simple_wcs)

        assert_allclose(c1[0], c2[0])
        assert_allclose(c1[1], c2[1])
        assert_allclose(s1, s2)


class TestGWCSDispatch:
    """
    Test that dispatch helpers correctly handle WCS objects without
    the ``has_distortion`` attribute (e.g., GWCS), defaulting to the
    Jacobian path.
    """

    @pytest.fixture
    def mock_gwcs(self, simple_wcs):
        """
        Create a mock WCS that wraps a simple WCS but has no
        has_distortion attribute, simulating GWCS behavior.
        """
        class MockGWCS:
            def __init__(self, real_wcs):
                self._wcs = real_wcs

            def world_to_pixel(self, *args, **kwargs):
                return self._wcs.world_to_pixel(*args, **kwargs)

            def pixel_to_world(self, *args, **kwargs):
                return self._wcs.pixel_to_world(*args, **kwargs)

        return MockGWCS(simple_wcs)

    def test_no_has_distortion_attr(self, mock_gwcs):
        """
        The mock should not have has_distortion.
        """
        assert not hasattr(mock_gwcs, 'has_distortion')

    def test_sky_to_pixel_scales_uses_jacobian(self, mock_gwcs):
        """
        Without has_distortion, should use the Jacobian path and still
        produce valid results.
        """
        pix_position, sw, sh, pangle = sky_to_pixel_scales(
            WCS_CENTER, mock_gwcs, 0.0)
        assert isinstance(pix_position, tuple)
        assert sw > 0
        assert sh > 0
        assert isinstance(pangle, Angle)

    def test_pixel_to_sky_scales_uses_jacobian(self, mock_gwcs):
        """
        Without has_distortion, pixel_to_sky_scales should use the
        Jacobian path.
        """
        xy_coord = (9.5, 9.5)
        sky_position, sw, sh, _sangle = pixel_to_sky_scales(
            xy_coord, mock_gwcs, 0.0)
        assert isinstance(sky_position, SkyCoord)
        assert sw > 0
        assert sh > 0

    def test_sky_to_pixel_mean_scale_uses_jacobian(self, mock_gwcs):
        """
        Without has_distortion, should use the Jacobian path.
        """
        pix_position, scale = sky_to_pixel_mean_scale(WCS_CENTER, mock_gwcs)
        assert isinstance(pix_position, tuple)
        assert scale > 0

    def test_pixel_to_sky_mean_scale_uses_jacobian(self, mock_gwcs):
        """
        Without has_distortion, should use the Jacobian path.
        """
        xy_coord = (9.5, 9.5)
        sky_position, scale = pixel_to_sky_mean_scale(xy_coord, mock_gwcs)
        assert isinstance(sky_position, SkyCoord)
        assert scale > 0

    def test_gwcs_scale_matches_simple(self, mock_gwcs, simple_wcs):
        """
        The mock GWCS (Jacobian path) should give scales close to the
        simple WCS (offset path).
        """
        _, scale_offset = sky_to_pixel_mean_scale(
            WCS_CENTER, simple_wcs)
        _, scale_jac = sky_to_pixel_mean_scale(
            WCS_CENTER, mock_gwcs)
        assert_allclose(scale_offset, scale_jac)


class TestNonsquarePixels:
    """
    Tests for WCS with non-square pixels to verify scale separation.
    """

    def test_nonsquare_mean_scale(self, nonsquare_wcs):
        """
        The mean scale should be near the arithmetic mean of the two
        singular values (1/cdelt_x and 1/cdelt_y in pix/arcsec).
        """
        _, scale = jacobian_sky_to_pixel_mean_scale(
            WCS_CENTER, nonsquare_wcs)
        # Arithmetic mean of singular values (SVD):
        # 1/cdelt_x_arcsec and 1/cdelt_y_arcsec
        cdelt_x = 0.03 * 3600
        cdelt_y = 0.05 * 3600
        expected = 0.5 * (1.0 / cdelt_x + 1.0 / cdelt_y)
        assert_allclose(scale, expected, rtol=1e-6)

    def test_nonsquare_directional_scales(self, nonsquare_wcs):
        """
        For non-square pixels, the directional scales should differ.
        """
        _, sw, sh, _ = jacobian_sky_to_pixel_scales(
            WCS_CENTER, nonsquare_wcs, 0.0)
        # Scale factors should differ since x and y pixel scales differ
        assert not np.isclose(sw, sh)


class TestSVDEllipseConversions:
    """
    Tests for `pixel_ellipse_to_sky_svd` and `sky_ellipse_to_pixel_svd`.
    """

    def test_pixel_to_sky_return_types(self, simple_wcs, center_xy_coord):
        """
        Should return (SkyCoord, float, float, Angle).
        """
        center, w, h, angle = pixel_ellipse_to_sky_svd(
            center_xy_coord, simple_wcs, 10.0, 5.0, 0.5)
        assert isinstance(center, SkyCoord)
        assert isinstance(w, (float, np.floating))
        assert isinstance(h, (float, np.floating))
        assert isinstance(angle, Angle)

    def test_sky_to_pixel_return_types(self, simple_wcs):
        """
        Should return (tuple, float, float, Angle).
        """
        center, w, h, angle = sky_ellipse_to_pixel_svd(
            WCS_CENTER, simple_wcs, 36.0, 18.0, 0.5)
        assert isinstance(center, tuple)
        assert isinstance(w, (float, np.floating))
        assert isinstance(h, (float, np.floating))
        assert isinstance(angle, Angle)

    def test_roundtrip_sky_pixel_sky(self, simple_wcs):
        """
        Sky -> pixel -> sky should recover the original ellipse.
        """
        sky_w, sky_h, sky_a = 36.0, 18.0, 0.5
        center_pix, pw, ph, pa = sky_ellipse_to_pixel_svd(
            WCS_CENTER, simple_wcs, sky_w, sky_h, sky_a)
        _, rw, rh, ra = pixel_ellipse_to_sky_svd(
            center_pix, simple_wcs, pw, ph, pa.rad)
        assert_allclose(rw, sky_w, rtol=1e-6)
        assert_allclose(rh, sky_h, rtol=1e-6)
        assert_allclose(ra.rad, sky_a, rtol=1e-4)

    def test_roundtrip_pixel_sky_pixel(self, simple_wcs, center_xy_coord):
        """
        Pixel -> sky -> pixel should recover the original ellipse.
        """
        pix_w, pix_h, pix_a = 10.0, 5.0, 0.3
        _, sw, sh, sa = pixel_ellipse_to_sky_svd(
            center_xy_coord, simple_wcs, pix_w, pix_h, pix_a)
        _, rw, rh, ra = sky_ellipse_to_pixel_svd(
            WCS_CENTER, simple_wcs, sw, sh, sa.rad)
        assert_allclose(rw, pix_w, rtol=1e-6)
        assert_allclose(rh, pix_h, rtol=1e-6)
        assert_allclose(ra.rad, pix_a, rtol=1e-4)

    def test_simple_wcs_width_height_scale(self, simple_wcs, center_xy_coord):
        """
        For a simple WCS, pixel dimensions should scale by
        WCS_CDELT_ARCSEC.
        """
        pix_w, pix_h = 10.0, 5.0
        _, sw, sh, _ = pixel_ellipse_to_sky_svd(
            center_xy_coord, simple_wcs, pix_w, pix_h, 0.0)
        assert_allclose(sw, pix_w * WCS_CDELT_ARCSEC, rtol=1e-5)
        assert_allclose(sh, pix_h * WCS_CDELT_ARCSEC, rtol=1e-5)

    def test_height_larger_than_width(self, simple_wcs, center_xy_coord):
        """
        When height > width, the SVD should still correctly assign
        widths and heights.
        """
        pix_w, pix_h = 5.0, 10.0
        _, sw, sh, _ = pixel_ellipse_to_sky_svd(
            center_xy_coord, simple_wcs, pix_w, pix_h, 0.0)
        # Width should be smaller than height in sky coords too
        assert sw < sh

    def test_sip_wcs_positive_sizes(self, sip_wcs):
        """
        Sizes should be positive for distorted WCS.
        """
        xy_coord = (9.5, 9.5)
        _, sw, sh, _ = pixel_ellipse_to_sky_svd(
            xy_coord, sip_wcs, 8.0, 4.0, 0.0)
        assert sw > 0
        assert sh > 0

    def test_sip_wcs_roundtrip(self, sip_wcs):
        """
        Roundtrip with SIP WCS should recover the original ellipse.
        """
        sky_w, sky_h, sky_a = 0.36, 0.18, 0.7
        center_pix, pw, ph, pa = sky_ellipse_to_pixel_svd(
            WCS_CENTER, sip_wcs, sky_w, sky_h, sky_a)
        _, rw, rh, ra = pixel_ellipse_to_sky_svd(
            center_pix, sip_wcs, pw, ph, pa.rad)
        assert_allclose(rw, sky_w, rtol=1e-5)
        assert_allclose(rh, sky_h, rtol=1e-5)
        assert_allclose(ra.rad, sky_a, rtol=1e-4)

    def test_angle_wrapped(self, simple_wcs, center_xy_coord):
        """
        The output angle should be in [0, 360) degrees.
        """
        _, _, _, angle = pixel_ellipse_to_sky_svd(
            center_xy_coord, simple_wcs, 10.0, 5.0, 0.5)
        assert 0.0 <= angle.deg < 360.0


class TestSVDScales:
    """
    Tests for `sky_to_pixel_svd_scales` and `pixel_to_sky_svd_scales`.
    """

    def test_sky_to_pixel_return_types(self, simple_wcs):
        """
        Should return (tuple, float, float, Angle).
        """
        center, smaj, smin, angle = sky_to_pixel_svd_scales(
            WCS_CENTER, simple_wcs)
        assert isinstance(center, tuple)
        assert isinstance(smaj, (float, np.floating))
        assert isinstance(smin, (float, np.floating))
        assert isinstance(angle, Angle)

    def test_pixel_to_sky_return_types(self, simple_wcs, center_xy_coord):
        """
        Should return (SkyCoord, float, float, Angle).
        """
        center, smaj, smin, angle = pixel_to_sky_svd_scales(
            center_xy_coord, simple_wcs)
        assert isinstance(center, SkyCoord)
        assert isinstance(smaj, (float, np.floating))
        assert isinstance(smin, (float, np.floating))
        assert isinstance(angle, Angle)

    def test_simple_wcs_isotropic(self, simple_wcs):
        """
        For an isotropic WCS, major and minor scales should be equal.
        """
        _, smaj, smin, _ = sky_to_pixel_svd_scales(
            WCS_CENTER, simple_wcs)
        expected = 1.0 / WCS_CDELT_ARCSEC
        assert_allclose(smaj, expected, rtol=1e-5)
        assert_allclose(smin, expected, rtol=1e-5)

    def test_pixel_to_sky_simple_scales(self, simple_wcs, center_xy_coord):
        """
        For an isotropic WCS, pixel-to-sky scales should equal
        WCS_CDELT_ARCSEC.
        """
        _, smaj, smin, _ = pixel_to_sky_svd_scales(
            center_xy_coord, simple_wcs)
        assert_allclose(smaj, WCS_CDELT_ARCSEC, rtol=1e-5)
        assert_allclose(smin, WCS_CDELT_ARCSEC, rtol=1e-5)

    def test_major_geq_minor(self, simple_wcs):
        """
        The major scale should always be >= minor scale (SVD ordering).
        """
        _, smaj, smin, _ = sky_to_pixel_svd_scales(
            WCS_CENTER, simple_wcs)
        assert smaj >= smin

    def test_nonsquare_different_scales(self, nonsquare_wcs):
        """
        For non-square pixels, major and minor scales should differ.
        """
        _, smaj, smin, _ = sky_to_pixel_svd_scales(
            WCS_CENTER, nonsquare_wcs)
        assert not np.isclose(smaj, smin)

    def test_roundtrip_inverse(self, simple_wcs, center_xy_coord):
        """
        The product of sky->pixel major scale and pixel->sky major scale
        should be ~1.
        """
        _, smaj_s2p, smin_s2p, _ = sky_to_pixel_svd_scales(
            WCS_CENTER, simple_wcs)
        _, smaj_p2s, smin_p2s, _ = pixel_to_sky_svd_scales(
            center_xy_coord, simple_wcs)
        assert_allclose(smaj_s2p * smaj_p2s, 1.0, rtol=1e-6)
        assert_allclose(smin_s2p * smin_p2s, 1.0, rtol=1e-6)

    def test_angle_wrapped(self, simple_wcs):
        """
        The output angle should be in [0, 360) degrees.
        """
        _, _, _, angle = sky_to_pixel_svd_scales(
            WCS_CENTER, simple_wcs)
        assert 0.0 <= angle.deg < 360.0

    def test_sip_wcs_positive_scales(self, sip_wcs):
        """
        Scales should be positive for distorted WCS.
        """
        _, smaj, smin, _ = sky_to_pixel_svd_scales(
            WCS_CENTER, sip_wcs)
        assert smaj > 0
        assert smin > 0
